import subprocess
import os


def lmp_lat_const(modelname):
    lmp_input_str = """# Define unit set and class of atomic model
  units metal
  atom_style atomic

  # Periodic boundary conditions along all three dimensions
  boundary p p p

  # lattice
  lattice   custom  3.2 &
            a1 1.0 0.0 0.0   a2 0.5 $(sqrt(3.0)/2.0) 0.0  a3 0.0 0.0 4.0 &
            basis 0.0 0.0 0.5 basis $(1.0/3.0) $(1.0/3.0) 0.375  basis $(1.0/3.0) $(1.0/3.0) 0.625 &
            spacing 1.0 $(sqrt(3.0)/2.0) 1.0  # lattice spacing is relative to lattice constant

  #this will create a tilted simbox
  region       simbox prism 0 1 0 1 0 4 0.5 0 0  units lattice  # (unit relative to lattice const )

  # create simbox that allows 2 types of atoms
  create_box     2 simbox

  # create atoms such that basis 1 is of type 1, and basis 2 and 3 are of type 2
  # the type (1) before box is not working here
  create_atoms   1 box basis 1 1 basis 2 2 basis 3 2

  mass      1 95.94
  mass      2 32.065

  # Specify which KIM Model to use, letting LAMMPS compute the virial/pressure
  pair_style    kim LAMMPSvirial rpls_modelname
  pair_coeff    * * Mo S

  # set pressure in x and y direction, so as to relax boxsize
  fix       1 all box/relax x 0.0 y 0.0

  # Set what information to write to dump file
  #dump      id all custom 1 lammps.dump id type x y z fx fy fz
  #dump_modify   id every 1 format "%d %d %13.5e %13.5e %13.5e %13.5e %13.5e %13.5e"

  # minimize energy
  minimize  1.0e-15 1.0e-15 10000 10000

  variable    mylx equal lx
  print "lat_const = ${mylx}"
  """

    # create lammps input file
    lmp_input_str = lmp_input_str.replace('rpls_modelname', modelname)
    with open('lammps.in', 'w') as fout:
        fout.write(lmp_input_str)

    # run lammps
    subprocess.call('lmp_serial <lammps.in > lammps.out', shell=True)

    # write results to edn format
    with open('lammps.out', 'r') as fin:
        for line in fin:
            if 'lat_const' in line:
                lat_const = float(line.split('=')[1])

    with open('lattice_const.edn', 'w') as fout:
        edn_str = '''{
      "species" {
        "source-value"  ["Mo" "Mo" "S"]
      }

      "lattice-const" {
        "source-unit" "Angstrom"
        "source-value"  %22.15e
      }
    }''' % (
            lat_const
        )

        fout.write(edn_str)


if __name__ == '__main__':
    lmp_lat_const()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from kliff.dataset import DataSet
from kliff.modelparameters import ModelParameters
from kliff.kimcalculator import KIMCalculator
from kliff.fisher import Fisher


# KIM model parameters
model = 'Three_Body_Stillinger_Weber_Si__MO_405512056662_004'
params = ModelParameters(model)
fname = '../tests/input/Si_SW_init_guess.txt'
params.read(fname)

# training set
tset = DataSet()
tset.read('../tests/training_set/Si_T300_4')
configs = tset.get_configurations()

# calculator
calc = KIMCalculator(model)
calc.create(configs)

# Fisher
fisher = Fisher(params, calc)
fim, fim_std = fisher.compute()

print('Fisher information:')
print(fim)
print('Stdandard deviation of Fisher information:')
print(fim_std)



import os
import numpy as np
import torch
import torch.nn as nn
import kliff

logger = kliff.logger.get_logger(__name__)


class ModelTorch(nn.Module):
    r"""Base class for machine learning models.

    Typically, a user will not directly use this.
    """

    def __init__(self, descriptor, seed=35):
        super(ModelTorch, self).__init__()

        self.seed = seed
        torch.manual_seed(seed)

        self.descriptor = descriptor
        dtype = self.descriptor.get_dtype()
        if dtype == np.float32:
            self.dtype = torch.float32
        elif dtype == np.float64:
            self.dtype = torch.float64
        else:
            raise ModelTorchError('Not support dtype "{}".'.format(dtype))

        self.save_prefix = None
        self.save_start = None
        self.save_frequency = None

    def forward(self, x):
        raise ModelTorchError('"forward" not implemented.')

    def fit(self, path):
        raise ModelTorchError(
            '"fit" not supported by this model. Minimize a loss function to train the '
            'model instead.'
        )

    def set_save_metadata(self, prefix, start, frequency):
        r"""Set metadata that controls how the model are saved during training.

        If this function is called before minimization starts, the model will be
        saved to the directory specified by ``prefix`` every ``frequency``
        epochs, beginning at the ``start`` epoch.

        Parameters
        ----------
        prefix: str
            Directory where the models are saved.
            Model will be named as `{prefix}/model_epoch{ep}.pt`, where `ep` is
            the epoch number.

        start: int
            Epoch number at which begins to save the model.

        frequency: int
            Save the model every ``frequency`` epochs.
        """
        self.save_prefix = str(prefix)
        self.save_start = int(start)
        self.save_frequency = int(frequency)

    def save(self, path):
        r"""Save a model to disk.

        Parameters
        ----------
        path: str
            Path where to store the model.
        """
        path = os.path.abspath(path)
        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        torch.save(self.state_dict(), path)

        # save descriptor mean and stdev
        fname = os.path.join(dirname, 'mean_and_stdev.pkl')
        self.descriptor.dump_mean_stdev(fname)

    def load(self, path, mode='train'):
        r"""Load a model on disk into memory.

        Parameters
        ----------
        path: str
            Path where the model is stored.

        mode: str
            Purpose of the loaded model. Should be either ``train`` or ``eval``.
        """

        self.load_state_dict(torch.load(path))
        if mode == 'train':
            self.train()
        elif mode == 'eval':
            self.eval()
        else:
            raise ModelTorchError('Unrecognized mode "{}" in model.load().'.format(mode))

        # load descriptor mean and stdev
        dirname = os.path.dirname(path)
        fname = os.path.join(dirname, 'mean_and_stdev.pkl')
        self.descriptor.load_mean_stdev(fname)

    def write_kim_model(self, path=None):
        raise ModelTorchError('"write_kim_model" not implemented.')


class ModelTorchError(Exception):
    def __init__(self, msg):
        super(ModelTorchError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg



import sys
import numpy as np
from collections import OrderedDict
import kliff
from .parameter import FittingParameter
from ..log import log_entry

logger = kliff.logger.get_logger(__name__)


class ComputeArguments:
    r"""Implementation of code to compute energy, forces, and stress.

    This is the base class for other compute arguments. Typically, a user will not
    directly use this.
    """

    implemented_property = []

    def __init__(
        self,
        conf,
        supported_species,
        influence_distance,
        compute_energy=True,
        compute_forces=True,
        compute_stress=False,
    ):
        self.conf = conf
        self.supported_species = supported_species
        self.influence_distance = influence_distance
        self.compute_energy = compute_energy
        self.compute_forces = compute_forces
        self.compute_stress = compute_stress
        self.compute_property = self.check_compute_property()
        self.results = dict([(i, None) for i in self.implemented_property])

    def refresh(self, influence_distance=None, params=None):
        r"""Refresh settings.

        Such as recreating the neighbor list due to the change of cutoff.
        """
        if influence_distance is not None:
            infl_dist = influence_distance
        else:
            try:
                infl_dist = params['influence_distance'].get_value()[0]
            except KeyError:
                report_error('"influence_distance" not provided by "params".')

        self.influence_distance = infl_dist

        # NOTE to be filled
        # create neighbor list based on `infl_dist`

    # TODO also check that the conf provide these properties
    def check_compute_property(self):
        def add_to_compute_property(compute_property, name):
            if name not in self.implemented_property:
                raise NotImplementedError('"{}" not implemented in model.'.format(name))
            compute_property.append(name)

        compute_property = []
        if self.compute_energy:
            add_to_compute_property(compute_property, 'energy')
        if self.compute_forces:
            add_to_compute_property(compute_property, 'forces')
        if self.compute_stress:
            add_to_compute_property(compute_property, 'stress')
        return compute_property

    def compute(self, params):
        r"""Compute the properties required by the compute flags, and store them in
        self.results.

        Parameters
        ----------
        params: dict of 1D array
            Parameters of the model.

        Example
        -------

        energy = a_func_to_compute_energy()
        forces = a_func_to_compute_forces()
        stress = a_func_to_compute_stress()
        self.results['energy'] = energy
        self.results['forces'] = forces
        self.results['stress'] = stress
        """
        # NOTE to be filled
        raise NotImplementedError('"compute" method of "ComputeArguments" not defined.')

    def get_compute_flag(self, name):
        if name in self.compute_property:
            return True
        else:
            return False

    def get_property(self, name):
        if name not in self.compute_property:
            report_error('Model not initialized to comptue "{}".'.format(name))
        result = self.results[name]
        if isinstance(result, np.ndarray):
            result = result.copy()
        return result

    def get_energy(self):
        return self.get_property('energy')

    def get_forces(self):
        return self.get_property('forces')

    def get_stress(self):
        return self.get_property('stress')

    def get_prediction(self):
        if self.compute_energy:
            energy = self.results['energy']
            pred = np.asarray([energy])
        else:
            pred = np.asarray([])

        if self.compute_forces:
            forces = self.results['forces']
            pred = np.concatenate((pred, forces.ravel()))

        if self.compute_stress:
            stress = self.results['stress']
            pred = np.concatenate((pred, stress))

        return pred

    def get_reference(self):
        if self.compute_energy:
            energy = self.conf.energy
            ref = np.asarray([energy])
        else:
            ref = np.asarray([])

        if self.compute_forces:
            forces = self.conf.forces
            ref = np.concatenate((ref, forces.ravel()))

        if self.compute_stress:
            stress = self.conf.stress
            ref = np.concatenate((ref, stress))

        return ref


class Model:
    r"""Base class for all physics-motivated models.

    Typically, a user will not directly use this.

    Parameters
    ----------
    model_name: str (optional)
        Name of the model.

    params_relation_callback: function (optional)
        A callback function to set the relations between parameters, which are
        called each minimization step after the optimizer updates the
        parameters.

        The function with be given the ModelParameters instance as argument, and
        it can use get_value() and set_value() to manipulate the relation
        between parameters.

        Example
        -------

        In the following example, we set the value of B[0] to 2 * A[0].

        def params_relation(params):
            A = params.get_value('A')
            B = params.get_value('B')
            B[0] = 2 * A[0]
            params.set_value('B', B)
    """

    def __init__(self, model_name=None, params_relation_callback=None):
        # TODO params_relation_callbacks may not work for some minimization
        # algorithms since abruptly change the parameter relation may result in
        # loss to go up, then the optimization method can fail. So to a check to
        # implement this only for the minimization methods that support it
        # natively.
        # Update: 2019/07/19
        # This should be fine as long as the dependent parameter is not in the `fitting
        # parameters` vector passed to an optimizer. The dependent parameter can just be
        # thought as a implicit parameter.

        self.model_name = model_name
        if self.model_name is not None:
            self.model_name = self.model_name.rstrip('/')
        self.params_relation_callback = params_relation_callback

        # NOTE to be filled
        self.params = OrderedDict()
        # set up parameters of the calculator
        # e.g.
        # self.params['sigma'] = Parameter(0.5)
        # self.params['epsilon'] = Parameter(0.4)

        # NOTE to be filled
        self.influence_distance = None

        # NOTE to be filled, should a dictionary
        # Key and value are species string and integer code, respectively.
        # if None, it supportes any species
        self.supported_species = None

        # NOTE to be filled
        self.compute_arguments_class = ComputeArguments

        # TODO maybe use metaclass to call this automatically after initialization
        # NOTE do not forget to call this in the subclass
        self.fitting_params = self.init_fitting_params(self.params)

    def write_kim_model(self, path=None):
        # NOTE fill this
        report_error('This model does not support writing to a KIM model.')

    def set_params_relation_callback(self, params_relation_callback):
        r"""Register a function to set the relation between parameters."""
        self.params_relation_callback = params_relation_callback

    def get_influence_distance(self):
        return self.influence_distance

    def get_supported_species(self):
        return self.supported_species

    def get_model_params(self, name):
        r"""Return a copy of the values of parameter.

        Parameters
        ----------
        name: string
            Name of the parameter.
        """
        if name in self.params:
            return self.params[name].get_value()
        else:
            report_error('"{}" is not a parameter of calculator.'.format(name))

    def set_model_params(self, name, value, check_shape=True):
        r"""Update the parameter values.

        Parameters
        ----------
        name: str
            Name of the parameter.

        value: list of floats
            Value of hte parameter.

        check_shape: bool
            If ``True``, check the shape of ``value``.
        """
        if name in self.params:
            self.params[name].set_value(value, check_shape)
        else:
            report_error('"{}" is not a parameter of the model.'.format(name))

    #    def save_model_params(self, path):
    #        params = dict()
    #        for i, j in self.params.items():
    #            v = j.value
    #            if isinstance(v, np.ndarray):
    #                v = v.tolist()
    #            params[i] = v
    #        with open(path, 'w') as fout:
    #            yaml.dump(params, fout, default_flow_style=False)
    #
    #    def load_model_params(self, path):
    #        with open(path, 'r') as fin:
    #            params = yaml.safe_load(fin)
    #        for key, value in params.items():
    #            self.set_model_params(key, value, check_shape=False)

    def echo_model_params(self, path=None):
        r"""Print the optimizable parameters.

        Parameters
        ----------
        path: str (optional)
            Path to print the information. if ``None``, print to stdout.
        """

        if path is not None:
            fout = open(path, 'w')
        else:
            fout = sys.stdout

        print(file=fout)
        print('#' + '=' * 80, file=fout)
        print('# Available parameters to optimize.')
        print(file=fout)
        if self.model_name is None:
            name = self.__class__.__name__
        else:
            name = self.model_name
        print('# Model:', name, file=fout)
        print('#' + '=' * 80, file=fout)
        print(file=fout)
        for name, p in self.params.items():
            print('name:', name, file=fout)
            print('value:', p.value, file=fout)
            print('size:', p.size, file=fout)
            print('dtype:', p.dtype, file=fout)
            print('description:', p.description, file=fout)
            print(file=fout)

        if path is not None:
            fout.close()

    def init_fitting_params(self, params):
        return FittingParameter(params)

    def read_fitting_params(self, path):
        self.fitting_params.read(path)

    def set_fitting_params(self, **kwargs):
        self.fitting_params.set(**kwargs)

    def set_one_fitting_param(self, name, settings):
        self.fitting_params.set_one(name, settings)

    #    def save_fitting_params(self, path):
    #        self.fitting_params.save(path)
    #
    #    def load_fitting_params(self, path):
    #        self.fitting_params.load(path)

    def echo_fitting_params(self, path=None):
        self.fitting_params.echo_params(path)

    def get_number_of_opt_params(self):
        return self.fitting_params.get_number_of_opt_params()

    def get_opt_params(self):
        return self.fitting_params.get_opt_params()

    def get_opt_param_name_value_and_indices(self, k):
        return self.fitting_params.get_opt_param_name_value_and_indices(k)

    def has_opt_params_bounds(self):
        return self.fitting_params.has_opt_params_bounds()

    def get_opt_params_bounds(self):
        return self.fitting_params.get_opt_params_bounds()

    def update_fitting_params(self, opt_params):
        r"""Update from optimizer to fitting params."""
        self.fitting_params.update_params(opt_params)

    # TODO if parameters relation set, remove the parameters from fitting params
    # or at least check it is not in the fitting params
    def apply_params_relation(self):
        r"""Force user-specified relation between parameters."""
        if self.params_relation_callback is not None:
            self.params_relation_callback(self.fitting_params)

    def update_model_params(self):
        r"""Update from fitting params to model params."""
        for name, attr in self.fitting_params.params.items():
            self.set_model_params(name, attr['value'], check_shape=False)

    def save(self, path):
        r"""Save a model to disk.

        Parameters
        ----------
        path: str
            Path where to store the model.
        """
        self.fitting_params.save(path)

    def load(self, path):
        r"""Load a model on disk into memory.

        Parameters
        ----------
        path: str
            Path where the model is stored.
        """
        self.fitting_params.load(path)
        self.update_model_params()


class ModelError(Exception):
    def __init__(self, msg):
        super(ModelError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg


def report_error(msg):
    log_entry(logger, msg, level='error')
    raise ModelError(msg)



from ..utils import check_torch_available
from .calculator import Calculator

__all__ = ['Calculator']

if check_torch_available():
    from .calculator_torch import CalculatorTorch, CalculatorTorchDDPCPU

    __all__.extend(['CalculatorTorch', 'CalculatorTorchDDPCPU'])



import torch
from torch.utils.data import Dataset
from ..descriptors.descriptor import load_fingerprints


class FingerprintsDataset(Dataset):
    r"""Atomic environment fingerprints dataset.

    Parameters
    ----------
    path: string
        Path to the fingerprints file.

    transform: callable (optional)
        Optional transform to be applied on a sample.
    """

    def __init__(self, path, transform=None):
        self.fp = load_fingerprints(path)
        self.transform = transform

    def __len__(self):
        return len(self.fp)

    def __getitem__(self, index):
        sample = self.fp[index]
        if self.transform:
            sample = self.transform(sample)
        return sample


def fingerprints_collate_fn(batch):
    r"""Convert a batch of samples into tensor.

    Unlike the default_collate_fn(), which stack samples in the batch (requiring each
    sample having the same dimension), this function does not do the stack.

    Parameters
    ----------
    batch: list
        A batch of samples.

    Returns
    -------
    tensor_batch: list
        Transform each sample into a tensor.
    """
    tensor_batch = []
    for i, sample in enumerate(batch):
        tensor_sample = {}
        for key, value in sample.items():
            if type(value).__module__ == 'numpy':
                value = torch.from_numpy(value)
            tensor_sample[key] = value
        tensor_batch.append(tensor_sample)

    return tensor_batch



import os
import sys
import warnings
import pickle
from collections import OrderedDict
from collections.abc import Iterable
import numpy as np
import kliff
from ..error import InputError

logger = kliff.logger.get_logger(__name__)


class Parameter:
    r"""Parameter class."""

    def __init__(self, value, dtype='double', description=None):
        self.set_value(value)
        self.dtype = dtype
        self.description = description

    def get_value(self):
        return self.value.copy()

    def get_size(self):
        return len(self.value)

    def get_dtype(self):
        return self.dtype

    def get_description(self):
        return self.description

    def set_value(self, value, check_shape=True):
        if check_shape:
            if isinstance(value, Iterable):
                if any(isinstance(i, Iterable) for i in value):
                    raise ParameterError('Parameter should be a 1D array.')
                self._set_val(value)
            else:
                raise ParameterError('Parameter should be a 1D array.')
        else:
            self._set_val(value)

    def _set_val(self, value):
        self.value = np.asarray(value)
        self.size = len(value)

    def to_string(self):
        s = 'value: {}\n'.format(np.array_str(self.value))
        s += 'size: {}\n'.format(self.size)
        s += 'dtype: {}\n'.format(self.dtype)
        s += 'description: {}\n'.format(self.description)
        return s


class FittingParameter:
    r"""Class of model parameters that will be optimized.

    It interacts with optimizer to provide initial guesses of parameters and receive
    updated parameters from the optimizer.

    Parameters
    ----------
    model_params: OrderDict
        All the parameters of a model(calculator).
    """

    def __init__(self, model_params):
        self.model_params = model_params

        # key: parameter name
        # values: {'size', 'value', 'use_default', 'fix', 'lower_bound', 'upper_bound'}
        self.params = OrderedDict()

        # components of params that are optimized
        self._index = []

    def read(self, fname):
        r"""Read the parameters that will be optimized. (Interface to user)

        Each parameter is a 1D array, and each component of the parameter array should be
        listed in a new line. Each line can contains 1, 2, or 3 elements.

        1st element: float or `DEFAULT`
            Initial guess of the parameter component. If `DEFAULT` (case insensitive), the
            value from the calculator is used as the initial guess.

        The 2nd and 3rd elements are optional. If 2 elements are provided:

        2nd element: `FIX` (case insensitive)
            If `FIX`, the corresponding component will not be optimized.

        If 3 elements are provided:

        2nd element: float or `INF` (case insensitive)
            Lower bound of the parameter. `INF` indicates that the lower bound is
            negative infinite, i.e. no lower bound is applied.

        3rd element: float or `INF` (case insensitive)
            Upper bound of the parameter. `INF` indicates that the upper bound is
            positive infinite, i.e. no upper bound is applied.

        An alternative is self.set().

        Parameters
        ----------
        fname: str
          name of file that includes the fitting parameters

        Example
        -------

        A
        DEFAULT
        1.1

        B
        DEFAULT FIX
        1.1     FIX

        C
        DEFAULT  0.1  INF
        1.0      INF  2.1
        2.0      FIX
        """

        with open(fname, 'r') as fin:
            lines = fin.readlines()
            lines = remove_comments(lines)
        num_line = 0
        while num_line < len(lines):
            line = lines[num_line].strip()
            num_line += 1
            if line in self.params:
                msg = 'file "{}", line {}. Parameter "{}" already set.'.format(
                    fname, num_line, line
                )
                warnings.warn(msg, category=Warning)
            if line not in self.model_params:
                raise InputError(
                    'file "{}", line {}. Parameter "{}" not supported.'.format(
                        fname, num_line, line
                    )
                )
            name = line
            size = self.model_params[name].get_size()
            settings = []
            for j in range(size):
                settings.append(lines[num_line].split())
                num_line += 1
            self.set_one(name, settings)

    def set(self, **kwargs):
        r"""Set the parameters that will be optimized. (Interface to user)

        One or more parameters can be set. Each argument is for one parameter, where the
        argument name is the parameter name, the value of the argument is the
        settings(including initial value, fix flag, lower bound, and upper bound).

        The value of the argument should be a list of list, where each inner list is for
        one component of the parameter, which can contain 1, 2, or 3 elements.  See
        self.read() for the options of the elements.

        Example
        -------
        instance.set(A=[['DEFAULT'], [2.0, 1.0, 3.0]], B=[[1.0, 'FIX'], [2.0, 'INF',
        3.0]])
        """
        for name, settings in kwargs.items():
            if name in self.params:
                msg = 'Parameter "{}" already set.'.format(name)
                warnings.warn(msg, category=Warning)
            if name not in self.model_params:
                raise InputError('Parameter "{}" not supported.'.format(name))
            self.set_one(name, settings)

    def set_one(self, name, settings):
        r"""Set one parameter that will be optimized.

        The name of the parameter should be given as the first entry of a list (or tuple),
        and then each data line should be given in in a list.

        Parameters
        ----------
        name: string
            name of a fitting parameter

        settings: list of list
            initial value, flag to fix a parameter, lower and upper bounds of a parameter

        Example
        -------
        name = 'param_A'
        settings = [['default', 0, 20], [2.0, 'fix'], [2.2, 'inf', 3.3]]
        instance.set_one(name, settings)
        """
        # TODO update Example in docs, see test_parameter.py for example
        size = self.model_params[name].get_size()
        if len(settings) != size:
            raise InputError(
                'Incorrect number of initial values for parameter "{}".'.format(name)
            )

        tmp_dict = {
            'size': size,
            'value': [None for _ in range(size)],
            'use_default': [False for _ in range(size)],
            'fix': [False for _ in range(size)],
            'lower_bound': [None for _ in range(size)],
            'upper_bound': [None for _ in range(size)],
        }
        self.params[name] = tmp_dict

        for j, line in enumerate(settings):
            num_items = len(line)
            if num_items == 1:
                self._read_1_item(name, j, line)
            elif num_items == 2:
                self._read_2_item(name, j, line)
            elif num_items == 3:
                self._read_3_item(name, j, line)
            else:
                raise InputError(
                    'More than 3 elements listed at data line '
                    '{} for parameter "{}".'.format(j + 1, name)
                )
            self._check_bounds(name)

        self._set_index(name)

    def echo_params(self, fname=None, print_size=True):
        r"""Print the optimizing parameters to stdout or file.

        Parameters
        ----------
        fname: str
            Name of the file to print the optimizing parameters. If None, printing to
            stdout.

        print_size: bool
            Flag to indicate whether print the size of parameter. Recall that a parameter
            may have one or more values.
        """

        if fname is not None:
            fout = open(fname, 'w')
        else:
            fout = sys.stdout

        print('#' + '=' * 80, file=fout)
        print('# Model parameters that are optimized.', file=fout)
        print('#' + '=' * 80, file=fout)
        print(file=fout)

        for name, attr in self.params.items():
            if print_size:
                print(name, attr['size'], file=fout)
            else:
                print(name, file=fout)

            for i in range(attr['size']):
                print('{:24.16e}'.format(attr['value'][i]), end=' ', file=fout)

                if attr['fix'][i]:
                    print('fix', end=' ', file=fout)

                lb = attr['lower_bound'][i]
                ub = attr['upper_bound'][i]
                has_lb = lb is not None
                has_ub = ub is not None
                has_bounds = has_lb or has_ub
                if has_bounds:
                    if has_lb:
                        print('{:24.16e}'.format(lb), end=' ', file=fout)
                    else:
                        print('None', end=' ', file=fout)
                    if has_ub:
                        print('{:24.16e}'.format(ub), end=' ', file=fout)
                    else:
                        print('None', end=' ', file=fout)
                print(file=fout)

            print(file=fout)

        if fname is not None:
            fout.close()

    def save(self, path):
        path = os.path.abspath(path)
        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(path, 'wb') as f:
            pickle.dump(self.params, f)

    def load(self, path):
        # restore parameters
        with open(path, 'rb') as f:
            self.params = pickle.load(f)
        # restore index
        self._index = []
        for name in self.params.keys():
            self._set_index(name)

    def get_names(self):
        return self.params.keys()

    def get_size(self, name):
        return self.params[name]['size']

    def get_value(self, name):
        return self.params[name]['value'].copy()

    def set_value(self, name, value):
        self.params[name]['value'] = value

    def get_lower_bound(self, name):
        return self.params[name]['lower_bound'].copy()

    def get_upper_bound(self, name):
        return self.params[name]['upper_bound'].copy()

    def get_fix(self, name):
        return self.params[name]['fix'].copy()

    def get_number_of_opt_params(self):
        return len(self._index)

    def update_params(self, opt_x):
        r"""Update parameter values from optimizer. (Interface to optimizer)

        This is the opposite operation of get_opt_params().

        Parameters
        ----------
        opt_x: list of floats
            parameter values from the optimizer.
        """
        for k, val in enumerate(opt_x):
            name = self._index[k].name
            c_idx = self._index[k].c_idx
            self.params[name]['value'][c_idx] = val

    def get_opt_params(self):
        r"""Nest all parameter values (except the fix ones) to a list.

        This is the opposite operation of update_params(). This can be fed to the
        optimizer as the starting parameters.

        Return
        ------
            A list of nested optimizing parameter values.
        """
        opt_x0 = []
        for idx in self._index:
            name = idx.name
            c_idx = idx.c_idx
            opt_x0.append(self.params[name]['value'][c_idx])
        if len(opt_x0) == 0:
            raise ParameterError('No parameters specified to optimize.')
        return np.asarray(opt_x0)

    def get_opt_param_name_value_and_indices(self, k):
        r"""Get the `name`, `value`, `parameter_index`, and `component_index` of an
        optimizing parameter given the slot `k`.
        """
        name = self._index[k].name
        p_idx = self._index[k].p_idx
        c_idx = self._index[k].c_idx
        value = self.params[name]['value'][c_idx]
        return name, value, p_idx, c_idx

    def get_opt_params_bounds(self):
        r"""Get the lower and upper parameter bounds."""
        bounds = []
        for idx in self._index:
            name = idx.name
            c_idx = idx.c_idx
            lower = self.params[name]['lower_bound'][c_idx]
            upper = self.params[name]['upper_bound'][c_idx]
            bounds.append([lower, upper])
        return bounds

    def has_opt_params_bounds(self):
        r"""Whether bounds are set for some parameters."""
        bounds = self.get_opt_params_bounds()
        for lb, up in bounds:
            if lb is not None or up is not None:
                return True
        return False

    def _read_1_item(self, name, j, line):
        self._read_1st_item(name, j, line[0])

    def _read_2_item(self, name, j, line):
        self._read_1st_item(name, j, line[0])
        if line[1].lower() == 'fix':
            self.params[name]['fix'][j] = True
        else:
            raise InputError('Data at line {} of {} corrupted.'.format(j + 1, name))

    def _read_3_item(self, name, j, line):
        self._read_1st_item(name, j, line[0])

        if (line[1] is not None) and (
            not (isinstance(line[1], str) and line[1].lower() == 'none')
        ):
            try:
                self.params[name]['lower_bound'][j] = float(line[1])
            except ValueError as e:
                raise InputError(
                    '{}.\nData at line {} of {} corrupted.'.format(e, j + 1, name)
                )

        if (line[2] is not None) and (
            not (isinstance(line[2], str) and line[2].lower() == 'none')
        ):
            try:
                self.params[name]['upper_bound'][j] = float(line[2])
            except ValueError as e:
                raise InputError(
                    '{}.\nData at line {} of {} corrupted.'.format(e, j + 1, name)
                )

    def _read_1st_item(self, name, j, first):
        if isinstance(first, str) and first.lower() == 'default':
            self.params[name]['use_default'][j] = True
            model_value = self.model_params[name].get_value()
            self.params[name]['value'][j] = model_value[j]
        else:
            try:
                self.params[name]['value'][j] = float(first)
            except ValueError as e:
                raise InputError(
                    '{}.\nData at line {} of parameter "{}" corrupted.'.format(
                        e, j + 1, name
                    )
                )

    def _check_bounds(self, name):
        r"""Check whether the initial guess of a parameter is within its lower and
        upper bounds.
        """
        attr = self.params[name]
        for i in range(attr['size']):
            lower_bound = attr['lower_bound'][i]
            upper_bound = attr['upper_bound'][i]
            value = attr['value'][i]
            if lower_bound is not None:
                if value < lower_bound:
                    raise InputError(
                        'Initial guess at line {} of parameter "{}" '
                        'out of bounds.'.format(i + 1, name)
                    )
            if upper_bound is not None:
                if value > upper_bound:
                    raise InputError(
                        'Initial guess at line {} of parameter "{}" '
                        'out of bounds.'.format(i + 1, name)
                    )

    def _set_index(self, name):
        r"""Check whether a parameter component will be optimized or not (by checking its
        'fix' attribute). If yes, include it in the index list.

        Given a parameter and its values such as:

        PARAM_FREE_B
        1.1
        2.2  fix
        4.4  3.3  5.5

        the first slot (1.1) and the third slot (4.4) will be included in self._index,
        and later be optimized.
        """

        # TODO check if there is alternative so as not to use OrderedDict
        p_idx = list(self.model_params.keys()).index(name)
        size = self.params[name]['size']
        fix = self.params[name]['fix']
        for c_idx in range(size):
            if not fix[c_idx]:
                idx = Index(name, p_idx, c_idx)
                # check whether already in self._index
                already_in = False
                for k, i in enumerate(self._index):
                    if idx == i:
                        already_in = k
                        break
                if already_in is not False:
                    self._index[already_in] = idx
                else:
                    self._index.append(idx)


class Index(object):
    def __init__(self, name, parameter_index=None, component_index=None):
        self.name = name
        self.parameter_index = self.p_idx = parameter_index
        self.component_index = self.c_idx = component_index

    def set_parameter_index(self, index):
        self.parameter_index = self.p_idx = index

    def set_component_index(self, index):
        self.component_index = self.c_idx = index

    def __expr__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ != other.__dict__
        else:
            return True


def remove_comments(lines):
    r"""Remove lines in a string list that start with # and content after #."""
    processed_lines = []
    for line in lines:
        line = line.strip()
        if not line or line[0] == '#':
            continue
        if '#' in line:
            line = line[0 : line.index('#')]
        processed_lines.append(line)
    return processed_lines


class ParameterError(Exception):
    def __init__(self, msg):
        super(ParameterError, self).__init__(msg)
        self.msg = msg

    def __str__(self):
        return self.msg



import os
import logging
import warnings


class Logger:
    def __init__(self, level='warning', filename='kliff.log'):

        # remove log file if existing
        if os.path.exists(filename):
            os.remove(filename)

        logging.basicConfig(
            filename=filename, format='%(asctime)s:%(name)s:%(levelname)s: %(message)s'
        )

        self.level = level.upper()
        self.filename = filename
        self.loggers = []

    def set_level(self, level):
        self.level = level.upper()
        # reset level for previous instantiated loggers
        for l in self.loggers:
            l.setLevel(self.level)

    def get_logger(self, name=None):
        if name is None:
            logger = logging.getLogger()
        else:
            logger = logging.getLogger(name)
        logger.setLevel(self.level)

        self.loggers.append(logger)
        return logger


def log_entry(logger, message, level='info', print_end='\n', warning_category=Warning):
    logger_level = getattr(logger, level)
    logger_level(message)

    if level == 'info':
        print(message, end=print_end)
    elif level == 'warning':
        warnings.warn(message, category=warning_category)



import sys
import numpy as np
import pytest
from kliff.dataset import Dataset
from kliff.calculators import Calculator
from kliff.models import KIM

ref_energies = [-277.409737571, -275.597759276, -276.528342759, -275.482988187]


ref_forces = [
    [
        [-1.15948917e-14, -1.15948917e-14, -1.16018306e-14],
        [-1.16989751e-14, -3.93232367e-15, -3.93232367e-15],
        [-3.92538477e-15, -1.17128529e-14, -3.92538477e-15],
    ],
    [
        [1.2676956, 0.1687802, -0.83520474],
        [-1.2536342, 1.19394052, -0.75371034],
        [-0.91847129, -0.26861574, -0.49488973],
    ],
    [
        [0.18042083, -0.11940541, 0.01800594],
        [0.50030564, 0.09165797, -0.09694234],
        [-0.23992404, -0.43625564, 0.14952855],
    ],
    [
        [-1.1114163, 0.21071302, 0.55246303],
        [1.51947195, -1.21522541, 0.7844481],
        [0.54684859, 0.01018317, 0.5062204],
    ],
]


def test_main():

    # training set
    tset = Dataset()
    tset.read('./configs_extxyz/Si_4')
    configs = tset.get_configs()

    # model
    modelname = 'SW_StillingerWeber_1985_Si__MO_405512056662_005'
    model = KIM(modelname)

    # calculator
    calc = Calculator(model)
    compute_arguments = calc.create(configs)

    for i, ca in enumerate(compute_arguments):
        calc.compute(ca)
        energy = calc.get_energy(ca)
        forces = calc.get_forces(ca)[:3]

        assert energy == pytest.approx(ref_energies[i], 1e-6)
        assert np.allclose(forces, ref_forces[i])

    # Cannot set them all by calling this function only once, because the assertion
    # depends on order
    model.set_fitting_params(sigma=[['default']])
    model.set_fitting_params(A=[['default', 'fix']])
    model.set_fitting_params(B=[['default']])

    # update params
    x0 = calc.get_opt_params()
    x1 = [i + 0.1 for i in x0]
    calc.update_opt_params(x1)
    params = model.inquire_params()
    assert np.allclose(params['sigma'].get_value(), [x1[0]])
    assert np.allclose(params['B'].get_value(), [x1[1]])
    # restore params
    calc.update_opt_params(x0)


if __name__ == '__main__':
    test_main()



class InputError(Exception):
    def __init__(self, msg):
        super(InputError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg


class KeyNotFoundError(Exception):
    def __init__(self, msg):
        super(KeyNotFoundError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg


class SupportError(Exception):
    def __init__(self, msg):
        super(SupportError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg


class InitializationError(Exception):
    def __init__(self, value):
        self.value = value

    def __expr__(self):
        return repr(self.value) + ' initialization failed'


def report_import_error(name):
    raise ImportError(
        'Optional module "{}" needed by "KLIFF" not found. Please install it '
        'manually.'.format(name)
    )



import numpy as np
from ..error import InputError, KeyNotFoundError


def read_extxyz(fname):
    r"""Read atomic configuration stored in extended xyz format.

    Parameters
    ----------
    fname: str
        name of the extended xyz file

    Returns
    -------
    cell: 2D array of shape(3,3)
        supercell lattice vectors

    PBC: list of 3 bool
        periodic boundary conditions

    species: list of N str, where N is the number of atoms
        species of atoms

    coords: 2D array of shape (N, 3)
        coordinates of atoms

    energy: float (or None if not provided in file)
        potential energy of the configuration

    forces: 2D array of shape (N, 3) (or None if not provided in file)
        forces on atoms

    stress: list of 6 float (or None if not provided in file)
        stress on the cell in Voigt notation
    """
    with open(fname, 'r') as fin:
        lines = fin.readlines()

        try:
            natoms = int(lines[0].split()[0])
        except ValueError as e:
            raise InputError(
                '{}.\nCorrupted data at line 1 of file "{}".'.format(e, fname)
            )

        # lattice vector, PBC, energy, and stress
        line = lines[1].replace("'", '"')
        cell = parse_key_value(line, 'Lattice', 'float', 9, fname)
        cell = np.reshape(cell, (3, 3))
        PBC = parse_key_value(line, 'PBC', 'int', 3, fname)
        # energy is optional
        try:
            in_quotes = check_in_quotes(line, 'Energy', fname)
            energy = parse_key_value(line, 'Energy', 'float', 1, fname, in_quotes)[0]
        except KeyNotFoundError:
            energy = None
        # stress is optional
        try:
            stress = parse_key_value(line, 'Stress', 'float', 6, fname)
        except KeyNotFoundError:
            stress = None

        # body, species symbol, x, y, z (and fx, fy, fz if provided)
        species = []
        coords = []
        forces = []
        # is forces provided
        line = lines[2].strip().split()
        if len(line) == 4:
            has_forces = False
        elif len(line) == 7:
            has_forces = True
        else:
            raise InputError('Corrupted data at line 3 of file "{}" .'.format(fname))

        try:
            num_lines = 0
            for line in lines[2:]:
                num_lines += 1
                line = line.strip().split()
                if len(line) != 4 and len(line) != 7:
                    raise InputError(
                        'Corrupted data at line {} of file "{}".'.format(
                            num_lines + 3, fname
                        )
                    )
                if has_forces:
                    symbol, x, y, z, fx, fy, fz = line
                    species.append(symbol.lower().capitalize())
                    coords.append([float(x), float(y), float(z)])
                    forces.append([float(fx), float(fy), float(fz)])
                else:
                    symbol, x, y, z = line
                    species.append(symbol.lower().capitalize())
                    coords.append([float(x), float(y), float(z)])
        except ValueError as e:
            raise InputError(
                '{}.\nCorrupted data at line {} of file "{}".'.format(
                    e, num_lines + 3, fname
                )
            )

        if num_lines != natoms:
            raise InputError(
                'Corrupted data file "{}". Number of atoms is "{}", '
                'whereas number of data lines is "{}".'.format(fname, natoms, num_lines)
            )

        species = np.asarray(species)
        coords = np.asarray(coords)
        if has_forces:
            forces = np.asarray(forces)
        else:
            forces = None
        return cell, PBC, species, coords, energy, forces, stress


def write_extxyz(
    fname, cell, PBC, species, coords, energy=None, forces=None, stress=None
):
    r"""Write configuration info to a file in extended xyz format.

    Parameters
    ----------
    fname: str
        name of the written file

    cell: 2D array of shape(3,3)
        supercell lattice vectors

    PBC: list of 3 bool
        periodic boundary conditions

    species: list of N str, where N is the number of atoms
        species of atoms

    coords: 2D array of shape (N, 3)
        coordinates of atoms

    energy: float (optional)
        potential energy of the configuration

    forces: 2D array of shape (N, 3) (optional)
        forces on atoms

    stress: list of 6 float (optional)
        stress on the cell in Voigt notation
    """

    with open(fname, 'w') as fout:

        # first line (number of atoms)
        natoms = len(species)
        fout.write('{}\n'.format(natoms))

        # second line
        fout.write('Lattice="')
        for i, line in enumerate(cell):
            for j, item in enumerate(line):
                if i == 2 and j == 2:
                    fout.write('{:.15g}" '.format(item))
                else:
                    fout.write('{:.15g} '.format(item))

        PBC = [int(i) for i in PBC]
        fout.write('PBC="{} {} {}" '.format(PBC[0], PBC[1], PBC[2]))

        if energy is not None:
            fout.write('Energy="{:.15g}" '.format(energy))

        if stress is not None:
            fout.write('Stress="')
            for i, s in enumerate(stress):
                if i == 5:
                    fout.write('{:.15g}" '.format(s))
                else:
                    fout.write('{:.15g} '.format(s))

        properties = 'Properties=species:S:1:pos:R:3'
        if forces is not None:
            properties += ':for:R:3\n'
        else:
            properties += '\n'
        fout.write(properties)

        # body
        for i in range(natoms):
            fout.write('{:2s} '.format(species[i]))
            fout.write(
                '{:23.15e} {:23.15e} {:23.15e} '.format(
                    coords[i][0], coords[i][1], coords[i][2]
                )
            )
            if forces is not None:
                fout.write(
                    '{:23.15e} {:23.15e} {:23.15e}'.format(
                        forces[i][0], forces[i][1], forces[i][2]
                    )
                )
            fout.write('\n')


def check_key(line, key, fname):
    r"""Check whether a key or its lowercase counter part is in line."""
    if key not in line:
        key_lower = key.lower()
        if key_lower not in line:
            raise KeyNotFoundError(
                '"{}" not found at line 2 of file "{}".'.format(key, fname)
            )
        else:
            key = key_lower
    return key


def check_in_quotes(line, key, fname):
    r"""Check whether ``key=value`` or ``key="value"`` in line."""
    key = check_key(line, key, fname)
    value = line[line.index(key) :]
    value = value[value.index('=') + 1 :]
    value = value.lstrip(' ')
    if value[0] == '"':
        return True
    else:
        return False


def parse_key_value(line, key, dtype, size, fname, in_quotes=True):
    r"""Given key, parse a string like ``other stuff key="value" other stuff`` to get
    value.

    If there is not space in value, the quotes `"` can be omitted.

    Parameters
    ----------
    line: str
        The string line.

    key: str
        Keyword to parse.

    dtype: str
        Expected data type of value, `int` or `float`.

    size: int
        Expected size of value.

    fname: str
        File name where the line comes from.

    Return
    ------
    list
        Values associated with key.
    """
    line = line.strip()
    key = check_key(line, key, fname)
    try:
        value = line[line.index(key) :]
        if in_quotes:
            value = value[value.index('"') + 1 :]
            value = value[: value.index('"')]
        else:
            value = value[value.index('=') + 1 :]
            value = value.lstrip(' ')
            value += ' '  # add an whitespace at end in case this is the last key
            value = value[: value.index(' ')]
        value = value.split()
    except Exception as e:
        raise InputError(
            '{}.\nCorrupted "{}" data at line 2 of file "{}".'.format(e, key, fname)
        )

    if len(value) != size:
        raise InputError(
            'Incorrect size of "{}" at line 2 of file "{}";\n'
            'required: {}, provided: {}. Possibly, the quotes do not match.'.format(
                key, fname, size, len(value)
            )
        )
    try:
        if dtype == 'float':
            value = [float(i) for i in value]
        elif dtype == 'int':
            value = [int(i) for i in value]
    except Exception as e:
        raise InputError(
            '{}.\nCorrupted "{}" data at line 2 of file "{}".'.format(e, key, fname)
        )

    return np.asarray(value)



"""
.. _tut_linear_regression:

Train a linear regression potential
===================================

In this tutorial, we train a linear regression model on the descriptors obtained using the
symmetry functions.
"""

from kliff.descriptors import SymmetryFunction
from kliff.dataset import Dataset
from kliff.models import LinearRegression
from kliff.calculators import CalculatorTorch


descriptor = SymmetryFunction(
    cut_name='cos', cut_dists={'Si-Si': 5.0}, hyperparams='set30', normalize=True
)


model = LinearRegression(descriptor)

# training set
dataset_name = 'Si_training_set/varying_alat'
tset = Dataset()
tset.read(dataset_name)
configs = tset.get_configs()
print('Number of configurations:', len(configs))

# calculator
calc = CalculatorTorch(model)
calc.create(configs, reuse=True)


##########################################################################################
# We can train a linear regression model by minimizing a loss function as discussed in
# :ref:`tut_nn`. But linear regression model has analytic solutions, and thus we can train
# the model directly by using this feature. This can be achieved by calling the ``fit()``
# function of its calculator.
#

# fit the model
calc.fit()


# save model
model.save('linear_model.pkl')



import os
import multiprocessing as mp
import torch
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallelCPU
from torch.utils.data import DataLoader
import kliff
from ..dataset.dataset import Configuration
from ..dataset.dataset_torch import FingerprintsDataset, fingerprints_collate_fn


logger = kliff.logger.get_logger(__name__)


class CalculatorTorch:
    r"""A calculator based on PyTorch.

    Parameters
    ----------
    model: obj
        Instance of :class:`~kliff.neuralnetwork.NeuralNetwork`.

    """

    implemented_property = ['energy', 'forces', 'stress']

    def __init__(self, model):

        self.model = model
        self.dtype = self.model.descriptor.dtype
        self.fingerprints_path = None

        self.use_energy = None
        self.use_forces = None
        self.use_stress = None

        self.results = dict([(i, None) for i in self.implemented_property])

    def create(
        self,
        configs,
        use_energy=True,
        use_forces=True,
        use_stress=False,
        reuse=False,
        fingerprints_path=None,
        fingerprints_mean_and_stdev_path=None,
        serial=False,
        nprocs=mp.cpu_count(),
    ):
        r"""Process configs into fingerprints.

        Parameters
        ----------

        configs: list of Configuration object

        use_energy: bool (optional)
            Whether to require the calculator to compute energy.

        use_forces: bool (optional)
            Whether to require the calculator to compute forces.

        use_stress: bool (optional)
            Whether to require the calculator to compute stress.

        reuse: bool (optional)
            If ``True``, reuse the fingerprints if found existing one. If ``False``,
            generate fingerprints from scratch no matter there is existing one or not.

        fingerprints_path: string (optional)
            Path to the generated fingerprints. If ``None``, default to
            ``./fingerprint.pkl``.

        fingerprints_mean_and_stdev_path: string (optional)
            Path to the mean and standard deviation of the fingerprints. If ``normalize``
            is not required by a descriptor, this is ignored. Otherwise, the mean and
            standard deviation read from ``fingerprints_mean_and_stdev_path``, are used to
            normalize the fingerprints. If ``None``, mean and standard deviation will be
            calculated from the descriptors and write to
            ``./fingerprints_mean_and_stdev.pkl``;

        serial: bool (optional)
            Compute fingerprints in serial mode. Memory efficient.

        nprocs: int (optional)
            Number of processes to use. If ``serial`` is ``True``, this is ignored..
        """

        self.configs = configs
        self.use_energy = use_energy
        self.use_forces = use_forces
        self.use_stress = use_stress

        if isinstance(configs, Configuration):
            configs = [configs]

        # generate pickled fingerprints
        self.fingerprints_path = self.model.descriptor.generate_fingerprints(
            configs,
            use_forces,
            use_stress,
            reuse,
            fingerprints_path,
            fingerprints_mean_and_stdev_path,
            serial,
            nprocs,
        )

    def get_compute_arguments(self, batch_size=1):
        r"""Return a list of compute arguments, each associated with a configuration."""
        fname = self.fingerprints_path
        fp = FingerprintsDataset(fname)
        loader = DataLoader(
            dataset=fp, batch_size=batch_size, collate_fn=fingerprints_collate_fn
        )

        return loader

    def fit(self):
        path = self.fingerprints_path
        self.model.fit(path)

    def compute(self, batch):

        grad = self.use_forces or self.use_stress

        # collate batch input to NN
        zeta_config = [sample['zeta'] for sample in batch]
        if grad:
            for zeta in zeta_config:
                zeta.requires_grad_(True)

        # evaluate model (forward pass)
        zeta_stacked = torch.cat(zeta_config, dim=0)
        energy_atom = self.model(zeta_stacked)

        # energy
        natoms_config = [len(zeta) for zeta in zeta_config]
        energy_config = [e.sum() for e in torch.split(energy_atom, natoms_config)]

        # forces and stress (backward propagation)
        forces_config = []
        stress_config = []
        if grad:
            for i, sample in enumerate(batch):

                # derivative of energy w.r.t. zeta
                energy = energy_config[i]
                zeta = zeta_config[i]
                dedz = torch.autograd.grad(energy, zeta, create_graph=True)[0]
                zeta.requires_grad_(False)  # no need of grad any more

                if self.use_forces:
                    dzetadr_forces = sample['dzetadr_forces']
                    f = self.compute_forces(dedz, dzetadr_forces)
                    forces_config.append(f)

                if self.use_stress:
                    dzetadr_stress = sample['dzetadr_stress']
                    volume = sample['dzetadr_volume']
                    s = self.compute_stress(dedz, dzetadr_stress, volume)
                    stress_config.append(s)

        if not self.use_forces:
            forces_config = None
        if not self.use_stress:
            stress_config = None

        self.results['energy'] = energy_config
        self.results['forces'] = forces_config
        self.results['stress'] = stress_config
        return {'energy': energy_config, 'forces': forces_config, 'stress': stress_config}

    @staticmethod
    def compute_forces(denergy_dzeta, dzetadr):
        forces = -torch.tensordot(denergy_dzeta, dzetadr, dims=([0, 1], [0, 1]))
        return forces

    @staticmethod
    def compute_stress(denergy_dzeta, dzetadr, volume):
        forces = torch.tensordot(denergy_dzeta, dzetadr, dims=([0, 1], [0, 1])) / volume
        return forces

    def get_energy(self, batch):
        return self.results['energy']

    def get_forces(self, batch):
        return self.results['forces']

    def get_stress(self, batch):
        return self.results['stress']


class CalculatorTorchDDPCPU(CalculatorTorch):
    def __init__(self, model, rank, world_size):
        super(CalculatorTorchDDPCPU, self).__init__(model)
        self.set_up(rank, world_size)

    def set_up(self, rank, world_size):
        os.environ['MASTER_ADDR'] = 'localhost'
        os.environ['MASTER_PORT'] = '12355'
        dist.init_process_group('gloo', rank=rank, world_size=world_size)

    def clean_up(self):
        dist.destroy_process_group()

    def compute(self, batch):
        grad = self.use_forces

        # collate batch input to NN
        zeta_config = self._collate(batch, 'zeta')
        if grad:
            for zeta in zeta_config:
                zeta.requires_grad_(True)
        zeta_stacked = torch.cat(zeta_config, dim=0)

        # evaluate model
        model = DistributedDataParallelCPU(self.model)
        energy_atom = model(zeta_stacked)

        # energy
        natoms_config = [len(zeta) for zeta in zeta_config]
        energy_config = [e.sum() for e in torch.split(energy_atom, natoms_config)]

        # forces
        if grad:
            dzetadr_config = self._collate(batch, 'dzetadr')
            forces_config = self.compute_forces_config(
                energy_config, zeta_config, dzetadr_config
            )
            for zeta in zeta_config:
                zeta.requires_grad_(False)
            # zeta_stacked.requires_grad_(False)
        else:
            forces_config = None

        return {'energy': energy_config, 'forces': forces_config}

    def __del__(self):
        self.clean_up()



import numpy as np
from kimpy import neighlist as nl
from .atomic_data import atomic_number, atomic_species


class NeighborList:
    r"""Neighbor list class .

    This is based on kimpy.neighlist.

    This uses the same approach that `LAMMPS` and `KIM` adopt: The atoms in the
    configuration (assuming a total of N atoms) are named contributing atoms, and padding
    atoms are created to satisfy the boundary conditions. The contributing atoms are
    numbered as 1, 2, ... N-1, and the padding atoms are numbered as N, N+1, N+2...
    Neighbors of atom can include both contributing atoms and padding atoms.

    Parameters
    ----------
    conf: Configuration object
        It stores the atoms information.

    infl_dist: float
        Influence distance, within which atoms are interacting with each other. In
        literatures, this is usually referred as `cutoff`.

    padding_need_neigh: bool
        Whether to generate neighbors for padding atoms.

    Attributes
    ----------
    coords: 2D array
        Coordinates of contributing and padding atoms.

    species: list
        Species string of contributing and padding atoms.

    image: 1D array
        Atom index, of which an atom is an image. The image of a contributing atom is
        itself.

    padding_coords: 2D array
        Coordinates of padding atoms.

    padding_species: list
        Species string and padding atoms.

    padding_image: 1D array
        Atom index, of which a padding atom is an image.

    Note
    ----
    To get the total force on a contributing atom, the forces on all padding atoms who are
    images of the contributing atom should be added back to the contributing atom.
    """

    def __init__(self, conf, infl_dist, padding_need_neigh=False):
        self.conf = conf
        self.infl_dist = infl_dist
        self.padding_need_neigh = padding_need_neigh

        # all atoms: contrib + padding
        self.coords = None
        self.species = None
        self.image = None

        self.padding_coords = None
        self.padding_species = None
        # padding_image[0] = 3: padding atom 1 is the image of contributing atom 3
        self.padding_image = None

        # neigh
        self.neigh = nl.initialize()
        self.create_neigh()

    def create_neigh(self):
        coords_cb = np.asarray(self.conf.get_coordinates(), dtype=np.double)
        species_cb = self.conf.get_species()
        cell = np.asarray(self.conf.get_cell(), dtype=np.double)
        PBC = np.asarray(self.conf.get_PBC(), dtype=np.intc)

        # create padding atoms
        species_code_cb = np.asarray(
            [atomic_number[s] for s in species_cb], dtype=np.intc
        )
        out = nl.create_paddings(self.infl_dist, cell, PBC, coords_cb, species_code_cb)
        coords_pd, species_code_pd, image_pd, error = out
        check_error(error, 'nl.create_padding')
        species_pd = [atomic_species[i] for i in species_code_pd]

        self.padding_coords = np.asarray(coords_pd, dtype=np.double)
        self.padding_species = species_pd
        self.padding_image = np.asarray(image_pd, dtype=np.intc)

        num_cb = coords_cb.shape[0]
        num_pd = coords_pd.shape[0]

        self.coords = np.asarray(np.concatenate((coords_cb, coords_pd)), dtype=np.double)
        self.species = np.concatenate((species_cb, species_pd))
        self.image = np.asarray(
            np.concatenate((np.arange(num_cb), image_pd)), dtype=np.intc
        )
        # flag to indicate whether to create neighborlist for an atom
        need_neigh = np.ones(num_cb + num_pd, dtype=np.intc)
        if not self.padding_need_neigh:
            need_neigh[num_cb:] = 0

        # create neighbor list
        cutoffs = np.asarray([self.infl_dist], dtype=np.double)
        error = nl.build(self.neigh, self.coords, self.infl_dist, cutoffs, need_neigh)
        check_error(error, 'nl.build')

    def get_neigh(self, index):
        r"""Get the indices, coordinates, and species string of a given atom.

        Parameters
        ----------
        index: int
            Atom number whose neighbor info is requested.

        Returns
        -------
        neigh_indices: list
            Indices of neighbor atoms in self.coords and self.species.

        neigh_coords: 2D array
            Coordinates of neighbor atoms.

        neigh_species: list
            Species symbol of neighbor atoms.
        """

        cutoffs = np.asarray([self.infl_dist], dtype=np.double)
        neigh_list_index = 0
        num_neigh, neigh_indices, error = nl.get_neigh(
            self.neigh, cutoffs, neigh_list_index, index
        )
        check_error(error, 'nl.get_neigh')

        neigh_coords = self.coords[neigh_indices]
        neigh_species = self.species[neigh_indices]
        return neigh_indices, neigh_coords, neigh_species

    def get_numneigh_and_neighlist_1D(self, request_padding=False):
        r"""Get the number of neighbors and neighbor list for all atoms.

        Parameters
        ----------
        request_padding: bool
            If ``True``, the returned number of neighbors and neighbor list include
            those for padding atoms; If ``False``, only return these for contributing
            atoms.

        Return
        ------
        numneigh: 1D array
            Number of neighbors for all atoms.

        neighlist: 1D array
            Indices of the neighbors for all atoms stacked into a 1D array. Its total
            length is ``sum(numneigh)``, and the first ``numneigh[0]`` components are the
            neighbors of atom `0`, the next ``numneigh[1]`` components are the neighbors
            of atom `1` ....
        """
        if request_padding:
            if not self.padding_need_neigh:
                raise NeighborListError(
                    'Request to get neighbors of padding atoms, but '
                    '"padding_need_neigh" is set to "False" at initialization.'
                )
            N = len(self.coords)
        else:
            N = self.conf.get_number_of_atoms()

        cutoffs = np.asarray([self.infl_dist], dtype=np.double)
        neigh_list_index = 0

        numneigh = []
        neighlist = []
        for i in range(N):
            num_neigh, neigh_indices, error = nl.get_neigh(
                self.neigh, cutoffs, neigh_list_index, i
            )
            check_error(error, 'nl.get_neigh')
            numneigh.append(num_neigh)
            neighlist.append(neigh_indices)
        neighlist = np.asarray(np.concatenate(neighlist), dtype=np.intc)
        numneigh = np.asarray(numneigh, dtype=np.intc)

        return numneigh, neighlist

    def get_coords(self):
        r"""Return coords of both contributing and padding atoms."""
        return self.coords.copy()

    def get_species(self):
        r"""Return species of both contributing and padding atoms."""
        return self.species[:]

    def get_species_code(self, mapping):
        r"""Integer species code of both contributing and padding atoms.

        Parameters
        ----------
        mapping: dict
            A mapping between species string and its code.

        Return
        1D array
            Integer species code.
        """
        return np.asarray([mapping[s] for s in self.species], dtype=np.intc)

    def get_image(self):
        r"""Return image of both contributing and padding atoms."""
        return self.image.copy()

    def get_padding_coords(self):
        r"""Return coords of padding atoms."""
        return self.padding_coords.copy()

    def get_padding_speices(self):
        r"""Return species string of padding atoms."""
        return self.padding_speices[:]

    def get_padding_species_code(self, mapping):
        r"""Integer species code of padding atoms.

        Parameters
        ----------
        mapping: dict
            A mapping between species string and its code.

        Return
        ------
        1D array
            Integer species code.
        """
        return np.asarray([mapping[s] for s in self.padding_species], dtype=np.intc)

    def get_padding_image(self):
        r"""Return image of padding atoms."""
        return self.padding_image.copy()

    def __del__(self):
        nl.clean(self.neigh)


def assemble_forces(forces, n, padding_image):
    r"""Assemble forces on padding atoms back to contributing atoms.

    Parameters
    ----------
    forces: 2D array
      forces on both contributing and padding atoms

    n: int
      number of contributing atoms

    padding_image: 1D int array
      atom number, of which the padding atom is an image

    Return
    ------
      Total forces on contributing atoms.
    """

    # numpy slicing does not make a copy !!!
    total_forces = np.array(forces[:n])

    has_padding = True if padding_image.size != 0 else False

    if has_padding:

        pad_forces = forces[n:]
        n_padding = pad_forces.shape[0]

        if n < n_padding:
            for i in range(n):
                # indices of padding atoms that are images of contributing atom i
                indices = np.where(padding_image == i)
                total_forces[i] += np.sum(pad_forces[indices], axis=0)
        else:
            for f, org_index in zip(pad_forces, padding_image):
                total_forces[org_index] += f

    return total_forces


def assemble_stress(coords, forces, volume):
    r"""Calculate the stress using the negative f dot r method.

    Parameters
    ----------
    coords: 2D array
        Coordinates of both contributing and padding atoms.

    forces: 2D array
        Partial forces on both contributing and padding atoms.

    volume: float
        Volume of the configuration.
    """

    stress = np.zeros(6)
    stress[0] = -np.sum(np.multiply(coords[:, 0], forces[:, 0])) / volume
    stress[1] = -np.sum(np.multiply(coords[:, 1], forces[:, 1])) / volume
    stress[2] = -np.sum(np.multiply(coords[:, 2], forces[:, 2])) / volume
    stress[3] = -np.sum(np.multiply(coords[:, 1], forces[:, 2])) / volume
    stress[4] = -np.sum(np.multiply(coords[:, 2], forces[:, 0])) / volume
    stress[5] = -np.sum(np.multiply(coords[:, 0], forces[:, 1])) / volume

    return stress


class NeighborListError(Exception):
    def __init__(self, msg):
        super(NeighborListError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg


def check_error(error, message=None):
    if error != 0 and error is not None:
        raise NeighborListError('Calling "{}" failed.'.format(message))



import kliff
from kliff.dataset import DataSet
from kliff.calculators import KIM
from kliff.calculators import WrapperCalculator
from kliff.loss import Loss

kliff.logger.set_level('debug')


# training set
tset = DataSet()
tset.read('../tests/calculators/Si_T300_4/')
configs = tset.get_configs()

# calculators
calc1 = KIM(model_name='Three_Body_Stillinger_Weber_Si__MO_405512056662_004')
calc1.create(configs)
calc1.set_fitting_params(A=[[16.0, 1.0, 20]], B=[['DEFAULT']])
calc2 = KIM(model_name='Three_Body_Stillinger_Weber_Si__MO_405512056662_004')
calc2.create(configs)
calc2.set_fitting_params(sigma=[[2.0951, 'fix']], gamma=[[1.1]])

# wrapper
calc = WrapperCalculator(calc1, calc2)

# loss
with Loss(calc, nprocs=2) as loss:
    result = loss.minimize(method='L-BFGS-B', options={'disp': True, 'maxiter': 10})



import os
import numpy as np
import logging
from collections import OrderedDict
import kimpy
from kimpy import neighlist as nl
import kliff
from .model import ComputeArguments, Model
from .parameter import Parameter
from ..neighbor import assemble_forces, assemble_stress
from ..log import log_entry
from ..error import SupportError

logger = kliff.logger.get_logger(__name__)


class KIMComputeArguments(ComputeArguments):
    r"""KIM potentials arguments.

    Parameters
    ----------
    kim_ca: KIM compute argument

    supported_species: dict
        Key and value are species string and integer code, respectively.
    """

    implemented_property = []

    def __init__(self, kim_ca, *args, **kwargs):
        self.kim_ca = kim_ca
        super(KIMComputeArguments, self).__init__(*args, **kwargs)

        # neighbor list
        self.neigh = None
        self.num_contributing_particles = None
        self.num_padding_particles = None
        self.padding_image_of = None

        # model input
        self.num_particles = None
        self.species_code = None
        self.particle_contributing = None
        self.coords = None

        # model output
        self.energy = None
        self.forces = None

        self._init_neigh()

        self.refresh(self.influence_distance)

    def check_compute_property(self):
        def add_to_compute_property(compute_property, name):
            if name not in self.implemented_property:
                raise NotImplementedError(
                    '"{}" not implemented in calculator.'.format(name)
                )
            compute_property.append(name)

        self._get_implemented_property()
        compute_property = []
        if self.compute_energy:
            add_to_compute_property(compute_property, 'energy')
        if self.compute_forces:
            add_to_compute_property(compute_property, 'forces')
        if self.compute_stress:
            add_to_compute_property(compute_property, 'stress')
        return compute_property

    def _get_implemented_property(self):

        # check compute arguments
        kim_can = kimpy.compute_argument_name
        N = kim_can.get_number_of_compute_argument_names()

        for i in range(N):
            name, error = kim_can.get_compute_argument_name(i)
            check_error(error, 'kim_can.get_compute_argument_name')

            dtype, error = kim_can.get_compute_argument_data_type(name)
            check_error(error, 'kim_can.get_compute_argument_data_type')

            support_status, error = self.kim_ca.get_argument_support_status(name)
            check_error(error, 'kim_ca.get_argument_support_status')

            # calculator can only handle energy and forces
            if support_status == kimpy.support_status.required:
                if name != kim_can.partialEnergy and name != kim_can.partialForces:
                    report_error('Unsupported required ComputeArgument "{}"'.format(name))

            # supported property
            if name == kim_can.partialEnergy:
                self.implemented_property.append('energy')
            elif name == kim_can.partialForces:
                self.implemented_property.append('forces')
                self.implemented_property.append('stress')

        # check compute callbacks
        kim_ccn = kimpy.compute_callback_name
        num_callbacks = kim_ccn.get_number_of_compute_callback_names()

        for i in range(num_callbacks):
            name, error = kim_ccn.get_compute_callback_name(i)
            check_error(error, 'kim_ccn.get_compute_callback_name')

            support_status, error = self.kim_ca.get_callback_support_status(name)
            check_error(error, 'compute_arguments.get_callback_support_status')

            # calculator only provides get_neigh
            if support_status == kimpy.support_status.required:
                if name != kim_ccn.GetNeighborList:
                    report_error('Unsupported required ComputeCallback "{}"'.format(name))

    def refresh(self, influence_distance=None, params=None):
        self.influence_distance = influence_distance
        self.update_neigh(influence_distance)
        self.register_data(self.compute_energy, self.compute_forces)

    def update_neigh(self, influence_distance):
        r"""Update neighbor list and model input.

        Parameters
        ----------

        influence_distance: float
        """

        # inquire information from conf
        cell = np.asarray(self.conf.get_cell(), dtype=np.double)
        PBC = np.asarray(self.conf.get_PBC(), dtype=np.intc)
        contributing_coords = np.asarray(self.conf.get_coordinates(), dtype=np.double)
        contributing_species = self.conf.get_species()
        num_contributing = self.conf.get_number_of_atoms()
        self.num_contributing_particles = num_contributing

        # species support and code
        unique_species = list(set(contributing_species))
        species_map = dict()
        for s in unique_species:
            if s in self.supported_species:
                species_map[s] = self.supported_species[s]
            else:
                report_error('species "{}" not supported by model'.format(s))
        contributing_species_code = np.array(
            [species_map[s] for s in contributing_species], dtype=np.intc
        )

        if any(PBC):  # need padding atoms
            out = nl.create_paddings(
                influence_distance,
                cell,
                PBC,
                contributing_coords,
                contributing_species_code,
            )
            padding_coords, padding_species_code, self.padding_image_of, error = out
            check_error(error, 'nl.create_paddings')

            num_padding = padding_species_code.size
            self.num_particles = np.array([num_contributing + num_padding], dtype=np.intc)
            tmp = np.concatenate((contributing_coords, padding_coords))
            self.coords = np.asarray(tmp, dtype=np.double)
            tmp = np.concatenate((contributing_species_code, padding_species_code))
            self.species_code = np.asarray(tmp, dtype=np.intc)
            self.particle_contributing = np.ones(self.num_particles[0], dtype=np.intc)
            self.particle_contributing[num_contributing:] = 0
            # TODO check whether padding need neigh and create accordingly
            # for now, create neigh for all atoms, including paddings
            need_neigh = np.ones(self.num_particles[0], dtype=np.intc)

        else:  # do not need padding atoms
            self.padding_image_of = np.array([])
            self.num_particles = np.array([num_contributing], dtype=np.intc)
            self.coords = np.array(contributing_coords, dtype=np.double)
            self.species_code = np.array(contributing_species_code, dtype=np.intc)
            self.particle_contributing = np.ones(num_contributing, dtype=np.intc)
            need_neigh = self.particle_contributing

        error = nl.build(
            self.neigh,
            self.coords,
            influence_distance,
            np.asarray([influence_distance], dtype=np.double),
            need_neigh,
        )
        check_error(error, 'nl.build')

    def register_data(self, compute_energy=True, compute_forces=True):
        r"""Register model input and output data in KIM API."""

        # check whether model support energy and forces
        kim_can = kimpy.compute_argument_name
        if compute_energy:
            name = kim_can.partialEnergy
            support_status, error = self.kim_ca.get_argument_support_status(name)
            check_error(error, 'kim_ca.get_argument_support_status')
            if not (
                support_status == kimpy.support_status.required
                or support_status == kimpy.support_status.optional
            ):
                report_error('Energy not supported by model')

        if compute_forces:
            name = kim_can.partialForces
            support_status, error = self.kim_ca.get_argument_support_status(name)
            check_error(error, 'kim_ca.get_argument_support_status')
            if not (
                support_status == kimpy.support_status.required
                or support_status == kimpy.support_status.optional
            ):
                report_error('Forces not supported by model')

        # register argument
        error = self.kim_ca.set_argument_pointer(
            kim_can.numberOfParticles, self.num_particles
        )
        check_error(error, 'kim_can.set_argument_pointer')

        error = self.kim_ca.set_argument_pointer(
            kim_can.particleSpeciesCodes, self.species_code
        )
        check_error(error, 'kim_can.set_argument_pointer')

        error = self.kim_ca.set_argument_pointer(
            kim_can.particleContributing, self.particle_contributing
        )
        check_error(error, 'kim_can.set_argument_pointer')

        error = self.kim_ca.set_argument_pointer(kim_can.coordinates, self.coords)
        check_error(error, 'kim_can.set_argument_pointer')

        if compute_energy:
            self.energy = np.array([0.0], dtype=np.double)
            error = self.kim_ca.set_argument_pointer(kim_can.partialEnergy, self.energy)
            check_error(error, 'kim_can.set_argument_pointer')
        else:
            self.energy = None
            error = self.kim_ca.set_argument_null_pointer(kim_can.partialEnergy)
            check_error(error, 'kim_can.set_argument_null_pointer')

        if compute_forces:
            self.forces = np.zeros([self.num_particles[0], 3], dtype=np.double)
            error = self.kim_ca.set_argument_pointer(kim_can.partialForces, self.forces)
            check_error(error, 'kim_can.set_argument_pointer')
        else:
            self.forces = None
            error = self.kim_ca.set_argument_null_pointer(kim_can.partialForces)
            check_error(error, 'kim_can.set_argument_null_pointer')

    def compute(self, kim_model):
        error = kim_model.compute(self.kim_ca)
        check_error(error, 'kim_model.compute')

        if self.compute_energy:
            self.results['energy'] = self.energy[0]
        if self.compute_forces:
            forces = assemble_forces(
                self.forces, self.num_contributing_particles, self.padding_image_of
            )
            self.results['forces'] = forces
        if self.compute_stress:
            volume = self.conf.get_volume()
            stress = assemble_stress(self.coords, self.forces, volume)
            self.results['stress'] = stress

    def _init_neigh(self):

        # create neighborlist
        neigh = nl.initialize()
        self.neigh = neigh

        # register get neigh callback
        error = self.kim_ca.set_callback_pointer(
            kimpy.compute_callback_name.GetNeighborList, nl.get_neigh_kim(), neigh
        )
        check_error(error, 'compute_arguments.set_callback_pointer')

    def __del__(self):
        """Garbage collection to destroy the neighbor list automatically"""
        if self.neigh is not None:
            nl.clean(self.neigh)
            self.neigh = None


class KIM(Model):
    def __init__(self, model_name=None, params_relation_callback=None):
        super(KIM, self).__init__(model_name, params_relation_callback)

        if self.model_name is None:
            c = self.__class__.__name__
            raise KIMModelError(
                '"model_name" is a mandatory argument for the "{}" calculator.'.format(c)
            )

        self.kim_model = self._initialize()
        self.params = self.inquire_params()

        self.compute_arguments_class = KIMComputeArguments
        self.fitting_params = self.init_fitting_params(self.params)

        logger.info('"{}" instantiated.'.format(self.__class__.__name__))

    def _initialize(self):
        r"""Initialize the KIM object"""
        units_accepted, model, error = kimpy.model.create(
            kimpy.numbering.zeroBased,
            kimpy.length_unit.A,
            kimpy.energy_unit.eV,
            kimpy.charge_unit.e,
            kimpy.temperature_unit.K,
            kimpy.time_unit.ps,
            self.model_name,
        )
        check_error(error, 'kimpy.model.create')
        if not units_accepted:
            report_error('requested units not accepted in kimpy.model.create')
        return model

    def inquire_params(self):
        r"""Inquire the KIM model to get all parameters. """
        params = OrderedDict()

        num_params = self.kim_model.get_number_of_parameters()
        for i in range(num_params):
            out = self.kim_model.get_parameter_metadata(i)
            dtype, extent, name, description, error = out
            check_error(error, 'model.get_parameter_data_type_extent_and_description')

            values = []
            for j in range(extent):
                if str(dtype) == 'Double':
                    value, error = self.kim_model.get_parameter_double(i, j)
                    check_error(error, 'model.get_parameter_double')
                elif str(dtype) == 'Int':
                    value, error = self.kim_model.get_parameter_int(i, j)
                    check_error(error, 'model.get_parameter_int')
                else:  # should never reach here
                    report_error('get unexpeced parameter data type "{}"'.format(dtype))
                values.append(value)
                params[name] = Parameter(
                    value=values, dtype=dtype, description=description
                )
        return params

    def create_a_kim_compute_argument(self):
        kim_ca, error = self.kim_model.compute_arguments_create()
        check_error(error, 'kim_model.compute_arguments_create')
        return kim_ca

    def update_model_params(self):
        """ Update from fitting params to model params. """

        # only update optimizing components
        num_params = self.get_number_of_opt_params()
        for i in range(num_params):
            _, value, p_idx, c_idx = self.get_opt_param_name_value_and_indices(i)
            self.kim_model.set_parameter(p_idx, c_idx, value)

        # refresh model
        self.kim_model.clear_then_refresh()

        # TODO this seems unnecessary
        # the correct way is to reimplement set_model_param, get_model_param,
        # and echo_model_param. Also, inquire_model_params seems could be used as
        # get_model_params.
        # the consideration is that we need a parameters object to be passed to
        # FittingParams. This should not be a problem.
        # update model params of the model class
        for name, attr in self.fitting_params.params.items():
            self.set_model_params(name, attr['value'], check_shape=False)

        if logger.getEffectiveLevel() == logging.DEBUG:
            params = self.inquire_params()
            s = ''
            for name, p in params.items():
                s += '\nname: {}\n'.format(name)
                s += p.to_string()
            log_entry(logger, s, level='debug')

    def get_influence_distance(self):
        r"""Return the influence distance of a model."""
        return self.kim_model.get_influence_distance()

    def get_supported_species(self):
        r"""Get all the supported species.

        Return
        ------
        species: dict
            Key and value are species string and integer code, respectively.
        """
        species = {}
        num_kim_species = kimpy.species_name.get_number_of_species_names()

        for i in range(num_kim_species):
            species_name, error = kimpy.species_name.get_species_name(i)
            check_error(error, 'kimpy.species_name.get_species_name')
            supported, code, error = self.kim_model.get_species_support_and_code(
                species_name
            )
            check_error(error, 'kim_model.get_species_support_and_code')
            if supported:
                species[str(species_name)] = code

        return species

    def write_kim_model(self, path=None):
        r"""Write out a KIM model that can be used directly with the kim-api.

        This function typically write two files to `path`: (1) CMakeLists.txt, and (2)
        a parameter file like A.params. `path` will be created if it does not exist.

        Parameters
        ----------
        path: str (optional)
            Path to the newly trained model.  If `None`, it is set to
            `./MODEL_NAME_kliff_trained`, where `MODEL_NAME` is the `model_name` that
            provided at the instantiation of this class.

        Note
        ----
        This only works for parameterized KIM models that support the writing of
        parameters.
        """
        present, required, error = self.kim_model.is_routine_present(
            kimpy.model_routine_name.WriteParameterizedModel
        )
        check_error(error, 'kim_model.is_routine_is_routine_present')
        if not present:
            raise SupportError('This KIM model does not support writing parameters.')

        if path is None:
            model_name = self.model_name + '_kliff_trained'
            path = os.path.join(os.getcwd(), model_name)
        else:
            path = os.path.abspath(path)
            model_name = os.path.basename(path)
        if not os.path.exists(path):
            os.makedirs(path)

        error = self.kim_model.write_parameterized_model(path, model_name)
        check_error(error, 'kim_model.write_parameterized_model')

        msg = 'KLIFF trained model write to "{}"'.format(path)
        log_entry(logger, msg, level='info')


class KIMModelError(Exception):
    def __init__(self, msg):
        super(KIMModelError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg


def check_error(error, msg):
    if error != 0 and error is not None:
        msg = 'Calling "{}" failed.\nSee "kim.log" for more information.'.format(msg)
        log_entry(logger, msg, level='error')
        raise KIMModelError(msg)


def report_error(msg):
    log_entry(logger, msg, level='error')
    raise KIMModelError(msg)



from collections.abc import Iterable


def length_equal(a, b):
    if isinstance(a, Iterable) and isinstance(b, Iterable):
        if len(a) == len(b):
            return True
        else:
            return False
    # If one is Iterable and the other not, we treat them as equal because of broadcast
    else:
        return True


def check_torch_available():
    try:
        import torch

        return True
    except ImportError:
        return False


def split_string(string, length=80, starter=None):
    r"""Insert `\n` into long string such that each line has size no more than `length`.

    Parameters
    ----------
    length: int
        Targeted length of the each line.

    starter: string
        String to insert at the beginning of each line.

    """

    if starter is not None:
        target_end = length - len(starter) - 1
    else:
        target_end = length

    sub_string = []
    while string:
        end = target_end
        if len(string) > end:
            while end >= 0 and string[end] != ' ':
                end -= 1
            end += 1
        sub = string[:end].strip()
        if starter is not None:
            sub = starter + ' ' + sub
        sub_string.append(sub)
        string = string[end:]

    return '\n'.join(sub_string) + '\n'



from kliff.models import KIM
from kliff.calculators import Calculator
from kliff.dataset import Dataset
from kliff.analyzers import EnergyForcesRMSE


model = KIM(model_name='SW_StillingerWeber_1985_Si__MO_405512056662_005')
model.load('kliff_model.pkl')

tset = Dataset()
dataset_name = 'Si_training_set'
tset.read(dataset_name)
configs = tset.get_configs()

calc = Calculator(model)
calc.create(configs)

analyzer = EnergyForcesRMSE(calc)
analyzer.run(verbose=2, sort='energy')



import numpy as np
import kliff
from .model import ComputeArguments, Model
from .parameter import Parameter, ParameterError
from ..neighbor import NeighborList, assemble_forces, assemble_stress

logger = kliff.logger.get_logger(__name__)


class LJComputeArguments(ComputeArguments):
    r"""Lennard-Jones 6-12 potential model built within KLIFF."""

    implemented_property = ['energy', 'forces', 'stress']

    def __init__(self, *args, **kwargs):
        super(LJComputeArguments, self).__init__(*args, **kwargs)
        self.refresh(self.influence_distance)

    def refresh(self, influence_distance=None, params=None):
        """ Refresh settings.

        Recreating the neighbor list due to the change of influence distance.
        """
        if influence_distance is not None:
            infl_dist = influence_distance
        else:
            try:
                infl_dist = params['influence_distance'].get_value()[0]
            except KeyError:
                raise ParameterError('"influence_distance" not provided by calculator."')
        self.influence_distance = infl_dist

        # create neighbor list
        self.neigh = NeighborList(self.conf, infl_dist, padding_need_neigh=False)

    def compute(self, params):
        epsilon = params['epsilon'].get_value()[0]
        sigma = params['sigma'].get_value()[0]
        rcut = params['cutoff'].get_value()[0]
        coords = self.conf.coords

        coords_including_padding = self.neigh.coords
        if self.compute_forces:
            forces_including_padding = np.zeros_like(coords_including_padding)

        energy = 0
        for i, xyz_i in enumerate(coords):
            neighlist, _, _ = self.neigh.get_neigh(i)
            for j in neighlist:
                xyz_j = coords_including_padding[j]
                rij = xyz_j - xyz_i
                r = np.linalg.norm(rij)
                if self.compute_forces:
                    phi, dphi = self.calc_phi_dphi(epsilon, sigma, r, rcut)
                    energy += 0.5 * phi
                    pair = 0.5 * dphi / r * rij
                    forces_including_padding[i] = forces_including_padding[i] + pair
                    forces_including_padding[j] = forces_including_padding[j] - pair
                elif self.compute_energy:
                    phi = self.calc_phi(epsilon, sigma, r, rcut)
                    energy += 0.5 * phi

        if self.compute_energy:
            self.results['energy'] = energy
        if self.compute_forces:
            forces = assemble_forces(
                forces_including_padding, len(coords), self.neigh.padding_image
            )
            self.results['forces'] = forces
        if self.compute_stress:
            volume = self.conf.get_volume()
            stress = assemble_stress(
                coords_including_padding, forces_including_padding, volume
            )
            self.results['stress'] = stress

    @staticmethod
    def calc_phi(epsilon, sigma, r, rcut):
        if r > rcut:
            phi = 0
        else:
            sor = sigma / r
            sor6 = sor * sor * sor
            sor6 = sor6 * sor6
            sor12 = sor6 * sor6
            phi = 4 * epsilon * (sor12 - sor6)
        return phi

    @staticmethod
    def calc_phi_dphi(epsilon, sigma, r, rcut):
        if r > rcut:
            phi = 0.0
            dphi = 0.0
        else:
            sor = sigma / r
            sor6 = sor * sor * sor
            sor6 = sor6 * sor6
            sor12 = sor6 * sor6
            phi = 4 * epsilon * (sor12 - sor6)
            dphi = 24 * epsilon * (-2 * sor12 + sor6) / r
        return phi, dphi


class LennardJones(Model):
    def __init__(self, model_name=None, params_relation_callback=None):
        super(LennardJones, self).__init__(model_name, params_relation_callback)

        self.params['epsilon'] = Parameter(value=[1.0])
        self.params['sigma'] = Parameter(value=[2.0])
        self.params['cutoff'] = Parameter(value=[5.0])
        self.compute_arguments_class = LJComputeArguments
        self.fitting_params = self.init_fitting_params(self.params)

    def get_influence_distance(self):
        return self.params['cutoff'].get_value()[0]



from ..utils import check_torch_available
from .model import ComputeArguments
from .model import Model
from .lennard_jones import LennardJones
from .kim import KIM

__all__ = ['ComputeArguments', 'Model', 'LennardJones', 'KIM']

if check_torch_available():
    from .model_torch import ModelTorch
    from .neural_network import NeuralNetwork
    from .linear_regression import LinearRegression

    __all__.extend(['ModelTorch', 'NeuralNetwork', 'LinearRegression'])



import os


def update_version(version, path, key, in_quotes=False, extra_space=False):
    """Update version given in `key=version` structure."""
    with open(path, 'r') as fin:
        lines = fin.readlines()
    with open(path, 'w') as fout:
        for line in lines:
            if key in line:
                idx = line.index('=')
                line = line[: idx + 1]
                if extra_space:
                    line += ' '
                if in_quotes:
                    v = "'{}'".format(version)
                else:
                    v = '{}'.format(version)
                fout.write(line + v + '\n')
            else:
                fout.write(line)


if __name__ == '__main__':
    kliff_v = '0.1.4'
    kliff_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    path = os.path.join(kliff_dir, 'kliff', '__init__.py')
    update_version(kliff_v, path, '__version__', True, True)



import os
import numpy as np
import torch
import kliff
from .model_torch import ModelTorch
from ..log import log_entry

logger = kliff.logger.get_logger(__name__)


class NeuralNetwork(ModelTorch):
    r"""Neural Network model.

    A feed-forward neural network model.

    Parameters
    ----------
    descriptor: object
        A descriptor that transforms atomic environment information to the fingerprints,
        which are used as the input for the neural network.

    seed: int (optional)
        Global seed for random numbers.
    """

    def __init__(self, descriptor, seed=35):
        super(NeuralNetwork, self).__init__(descriptor, seed)

        self.layers = None

        logger.info('"{}" instantiated.'.format(self.__class__.__name__))

    def add_layers(self, *layers):
        r"""Add layers to the sequential model.

        Parameters
        ----------
        layers: torch.nn layers
            ``torch.nn`` layers that are used to build a sequential model.  Available ones
            including: torch.nn.Linear, torch.nn.Dropout, and torch.nn.Sigmoid among
            others. See https://pytorch.org/docs/stable/nn.html for a full list of
            torch.nn layers.
        """
        if self.layers is not None:
            report_error(
                '"add_layers" called multiple times. It should be called only once.'
            )
        else:
            self.layers = []

        for la in layers:
            self.layers.append(la)
            # set it as attr so that parameters are automatically registered
            setattr(self, 'layer_{}'.format(len(self.layers)), la)

        # check shape of first layer and last layer
        first = self.layers[0]
        if first.in_features != len(self.descriptor):
            report_error(
                '"in_features" of first layer should be equal to descriptor size.'
            )
        last = self.layers[-1]
        if last.out_features != 1:
            report_error('"out_features" of last layer should be 1.')

        # cast types
        self.type(self.dtype)

    def forward(self, x):
        for j, layer in enumerate(self.layers):
            x = layer(x)
        return x

    def write_kim_model(self, path=None, driver_name=None, dropout_ensemble_size=None):
        """Write out a model that is compatible with the KIM API.

        Parameters
        ----------
        path: string (optional)
            Path to write the model (i.e. the model name). If ``None``, defaults to
            ``NeuralNetwork_KLIFF__MO_000000111111_000``.

        driver_name: string (optional)
            Name of the model driver.

        dropout_ensemble_size: int (optional)
            Size of the dropout ensemble. Ignored if not fitting a dropout NN. If
            ``None``, defaults to 100.
        """

        if path is None:
            modelname = 'NeuralNetwork_KLIFF__MO_000000111111_000'
            path = os.path.join(os.getcwd(), modelname)
        else:
            path = os.path.abspath(path)
            modelname = os.path.basename(path)
        if not os.path.exists(path):
            os.makedirs(path)
        if driver_name is None:
            driver_name = 'DUNN_WT__MD_000000111111_000'

        desc_name = 'descriptor.params'
        nn_name = 'NN.params'
        dropout_name = 'dropout_binary.params'

        paramfiles = [desc_name, nn_name, dropout_name]
        self.write_kim_cmakelists(
            path, modelname, driver_name, paramfiles, version='2.0.0'
        )
        self.write_kim_params(path, nn_name)
        self.descriptor.write_kim_params(path, desc_name)
        self.write_kim_dropout_binary(path, dropout_name, dropout_ensemble_size)

        msg = 'KLIFF trained model write to "{}"'.format(path)
        log_entry(logger, msg, level='info')

    def write_kim_params(self, path, fname='NN.params'):

        weights, biases = self._get_weights_and_biases()
        activations = self._get_activations()
        drop_ratios = self._get_drop_ratios()

        # PyTorch uses x*W^T + b, so we need to transpose it.
        # see https://pytorch.org/docs/stable/nn.html#linear
        weights = [torch.t(w) for w in weights]

        with open(os.path.join(path, fname), 'w') as fout:
            # header
            fout.write('#' + '=' * 80 + '\n')
            fout.write(
                '# NN structure and parameters file generated by KLIFF\n'
                '# \n'
                '# Note that the NN assumes each row of the input "X" is an \n'
                '# observation, i.e. the layer is implemented as\n'
                '# Y = activation(XW + b).\n'
                '# You need to transpose your weight matrix if each column of "X" is \n'
                '# an observation.\n'
            )
            fout.write('#' + '=' * 80 + '\n\n')

            # number of layers
            num_layers = len(weights)
            fout.write(
                '{}    # number of layers (excluding input layer,including output '
                'layer)\n'.format(num_layers)
            )

            # size of layers
            for b in biases:
                fout.write('{}  '.format(len(b)))
            fout.write('  # size of each layer (last must be 1)\n')

            # activation function
            # TODO enable writing different activations for each layer
            activation = activations[0]
            fout.write('{}    # activation function\n'.format(activation))

            # keep probability
            for i in drop_ratios:
                fout.write('{:.15g}  '.format(1.0 - i))
            fout.write('  # keep probability of input for each layer\n\n')

            # weights and biases
            for i, (w, b) in enumerate(zip(weights, biases)):

                # weight
                rows, cols = w.shape
                if i != num_layers - 1:
                    fout.write(
                        '# weight of hidden layer {},  shape({}, {})\n'.format(
                            i + 1, rows, cols
                        )
                    )
                else:
                    fout.write(
                        '# weight of output layer, shape({}, {})\n'.format(rows, cols)
                    )
                for line in w:
                    for item in line:
                        if self.dtype == torch.float64:
                            fout.write('{:23.15e}'.format(item))
                        else:
                            fout.write('{:15.7e}'.format(item))
                    fout.write('\n')

                # bias
                if i != num_layers - 1:
                    fout.write(
                        '# bias of hidden layer {}, shape({}, )\n'.format(i + 1, cols)
                    )
                else:
                    fout.write('# bias of output layer, shape({}, )\n'.format(cols))
                for item in b:
                    if self.dtype == torch.float64:
                        fout.write('{:23.15e}'.format(item))
                    else:
                        fout.write('{:15.7e}'.format(item))
                fout.write('\n\n')

    def write_kim_dropout_binary(self, path, fname='dropout_binary.params', size=None):

        drop_ratios = self._get_drop_ratios()
        keep_prob = [1.0 - i for i in drop_ratios]
        _, biases = self._get_weights_and_biases()
        num_units = [self.descriptor.get_size()] + [len(i) for i in biases]

        no_drop = np.all(np.asarray(drop_ratios) < 1e-10)
        if no_drop:
            size = 0
        else:
            if size is None:
                size = 100

        with open(os.path.join(path, fname), 'w') as fout:
            fout.write('#' + '=' * 80 + '\n')
            fout.write(
                '# Dropout binary parameters file generated by KLIFF.\n'
                '#\n'
                '# Note, "ensemble size = 0", means that no dropout needs to be\n'
                '# applied at all.'
            )
            fout.write('#' + '=' * 80 + '\n\n')

            fout.write('{}  # ensemble size\n'.format(size))
            for rep in range(size):
                fout.write('#' + '=' * 80 + '\n')
                fout.write('# instance {}\n'.format(rep))
                for i in range(len(keep_prob)):
                    fout.write('# layer {}\n'.format(i))
                    n = num_units[i]
                    k = keep_prob[i]
                    rnd = np.floor(np.random.uniform(k, k + 1, n))
                    rnd = np.asarray(rnd, dtype=np.intc)
                    for d in rnd:
                        d = 1 if d > 1 else d
                        d = 0 if d < 0 else d
                        fout.write('{} '.format(d))
                    fout.write('\n')

    @staticmethod
    def write_kim_cmakelists(
        path,
        modelname='NeuralNetwork_KLIFF__MO_000000111111',
        driver_name='DUNN_WT__MD_000000111111_000',
        paramfiles=["descriptor.params", "NN.params", "dropout_binary.params"],
        version='2.0.0',
    ):
        with open(os.path.join(path, 'CMakeLists.txt'), 'w') as fout:
            fout.write('#\n')
            fout.write('# Contributors:\n')
            fout.write('#    KLIFF (https://kliff.readthedocs.io)\n')
            fout.write('#\n\n')
            fout.write('cmake_minimum_required(VERSION 3.4)\n\n')
            fout.write('list(APPEND CMAKE_PREFIX_PATH $ENV{KIM_API_CMAKE_PREFIX_DIR})\n')
            fout.write('find_package(KIM-API 2.0 REQUIRED CONFIG)\n')
            fout.write('if(NOT TARGET kim-api)\n')
            fout.write('  enable_testing()\n')
            fout.write(
                '  project("${KIM_API_PROJECT_NAME}" VERSION "${KIM_API_VERSION}"\n'
            )
            fout.write('    LANGUAGES CXX C Fortran)\n')
            fout.write('endif()\n\n')
            fout.write('add_kim_api_model_library(\n')
            fout.write('  NAME            "{}"\n'.format(modelname))
            fout.write('  DRIVER_NAME     "{}"\n'.format(driver_name))
            fout.write('  PARAMETER_FILES')
            for s in paramfiles:
                fout.write(' "{}"'.format(s))
            fout.write('\n')
            fout.write('  )\n')

    def _group_layers(
        self,
        param_layer=['Linear'],
        activ_layer=['Sigmoid', 'Tanh', 'ReLU', 'ELU'],
        dropout_layer=['Dropout'],
    ):
        r"""Divide all the layers into groups.

        The first group is either an empty list or a `Dropout` layer for the input layer.
        The last group typically contains only a `Linear` layer.  For other groups, each
        group contains two, or three layers. `Linear` layer and an activation layer are
        mandatory, and a third `Dropout` layer is optional.

        Return
        ------
        groups: list of list of layers
        """

        groups = []
        new_group = []

        supported = param_layer + activ_layer + dropout_layer
        for i, layer in enumerate(self.layers):
            name = layer.__class__.__name__
            if name not in supported:
                report_error(
                    'Layer "{}" not supported by KIM model. Cannot proceed '
                    'to write.'.format(name)
                )

            if name in activ_layer:
                if i == 0:
                    report_error('First layer cannot be a "{}" layer'.format(name))
                if self.layers[i - 1].__class__.__name__ not in param_layer:
                    report_error(
                        'Cannot convert to KIM model. a "{}" layer must follow '
                        'a "Linear" layer.'.format(name)
                    )
            if name[:7] in dropout_layer:
                if self.layers[i - 1].__class__.__name__ not in activ_layer:
                    report_error(
                        'Cannot convert to KIM model. a "{}" layer must follow '
                        'an activation layer.'.format(name)
                    )
            if name in param_layer:
                groups.append(new_group)
                new_group = []
            new_group.append(layer)
        groups.append(new_group)

        return groups, param_layer, activ_layer, dropout_layer

    def _get_weights_and_biases(self):
        r"""Get weights and biases of all layers that have weights and biases."""

        groups, supported, _, _ = self._group_layers()

        weights = []
        biases = []
        for i, g in enumerate(groups):
            if i != 0:
                layer = g[0]
                name = layer.__class__.__name__
                if name in supported:
                    weight = layer.weight
                    bias = layer.bias
                    weights.append(weight)
                    biases.append(bias)
        return weights, biases

    def _get_activations(self):
        r"""Get the activation of all layers."""

        groups, _, supported, _ = self._group_layers()

        activations = []
        for i, g in enumerate(groups):
            if i != 0 and i != (len(groups) - 1):
                layer = g[1]
                name = layer.__class__.__name__
                if name in supported:
                    activations.append(name.lower())
        return activations

    def _get_drop_ratios(self):
        r"""Get the dropout ratio of all layers."""

        groups, _, _, supported = self._group_layers()

        drop_ratios = []
        for i, g in enumerate(groups):
            if i == 0:
                if len(g) != 0:
                    layer = g[0]
                    name = layer.__class__.__name__
                    if name in supported:
                        drop_ratios.append(layer.p)
                else:
                    drop_ratios.append(0.0)
            elif i == len(groups) - 1:
                pass
            else:
                if len(g) == 3:
                    layer = g[2]
                    name = layer.__class__.__name__
                    if name in supported:
                        drop_ratios.append(layer.p)
                else:
                    drop_ratios.append(0.0)

        return drop_ratios


class NeuralNetworkError(Exception):
    def __init__(self, msg):
        super(NeuralNetworkError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg


def report_error(msg):
    log_entry(logger, msg, level='error')
    raise NeuralNetworkError(msg)



from kliff.descriptors import SymmetryFunction
from kliff.models import NeuralNetwork
from kliff import nn
from kliff.dataset import Dataset
from kliff.calculators import CalculatorTorch
from kliff.analyzers import EnergyForcesRMSE

# model
descriptor = SymmetryFunction(
    cut_name='cos', cut_dists={'Si-Si': 5.0}, hyperparams='set30', normalize=True
)

N1 = 10
N2 = 10
model = NeuralNetwork(descriptor)
model.add_layers(
    # first hidden layer
    nn.Linear(descriptor.get_size(), N1),
    nn.Tanh(),
    # second hidden layer
    nn.Linear(N1, N2),
    nn.Tanh(),
    # output layer
    nn.Linear(N2, 1),
)
model.load('./saved_model.pkl', mode='train')

# dataset
tset = Dataset()
dataset_name = 'Si_training_set'
tset.read(dataset_name)
configs = tset.get_configs()

# calculator
calc = CalculatorTorch(model)
calc.create(configs, reuse=True)

# analyzer
analyzer = EnergyForcesRMSE(calc)
analyzer.run(verbose=2, sort='energy')



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

# sys.path.insert(0, os.path.abspath('.'))

# this should be the path to the package, not the source
# e.g. ~/Applications/kliff
# NOTE, do not use sys.path.insert, especially true when you have a C extension.
# Because if you use `pip install .` or `python setup install` to build your
# extension and install the package, the C extension will not be placed in the
# sorce directory, but will be placed to your virtual environment. Then if you use
# sys.path.insert to insert your source directory to be the first place to look for
# your package, it will fail to find the C extension.
# Here, we add it for sphinx to find the package source package, in case we do not
# install the package.
sys.path.append(os.path.abspath('../../'))


# -- Project information -----------------------------------------------------

project = 'KLIFF'
copyright = '2019, Mingjian Wen'
author = 'Mingjian Wen'

# The short X.Y version
version = '0.0'
# The full version, including alpha/beta/rc tags
release = '0.0.1'


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.imgmath',
    # 'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    # 'sphinx.ext.todo',
    # 'sphinx.ext.coverage',
    'sphinx_gallery.gen_gallery',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = 'sphinx'
pygments_style = 'default'
# pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {'logo_only': True}  # shows only the logo with no title text

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = './img/logo.jpg'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'KLIFFdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'KLIFF.tex', 'KLIFF Documentation', 'Mingjian Wen', 'manual')
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'kliff', 'KLIFF Documentation', [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        'KLIFF',
        'KLIFF Documentation',
        author,
        'KLIFF',
        'One line description of project.',
        'Miscellaneous',
    )
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Mock setup --------------------------------------------------------------
# A list of modules to be mocked up.
# https://docs.readthedocs.io/en/stable/faq.html#i-get-import-errors-on-libraries-that-depend-on-c-modules
# This is useful when some external dependencies are not installed at build time,
# which can break the building process.
autodoc_mock_imports = [
    'numpy',
    'scipy',
    'scipy.optimize',
    'kimpy',
    'yaml',
    'ase',
    'torch',
    'tensorflow',
]

# do not sort member functions of a class
autodoc_member_order = 'bysource'

# -- Math setup --------------------------------------------------------------
# imgmath settings
imgmath_image_format = 'svg'
imgmath_latex_preamble = '\\usepackage{bm} \\usepackage{amsmath}'


# -- sphinx-galley setup ------------------------------------------------------
sphinx_gallery_conf = {
    'examples_dirs': '../../examples',  # path to your example scripts
    'gallery_dirs': 'auto_examples',  # path where to save gallery generated examples
    'filename_pattern': '/example_',  # include files with names 'example_'; '/' matches the directory name separator
    'ignore_pattern': '^((?!/example_).)*$',  # exclude files without name '/example_'
}



import os
import numpy as np
import scipy.optimize
import kliff
from . import parallel
from .log import log_entry
from .error import InputError, report_import_error

try:
    import torch

    torch_available = True
except ImportError:
    torch_available = False

try:
    from mpi4py import MPI

    mpi4py_available = True
except ImportError:
    mpi4py_available = False


logger = kliff.logger.get_logger(__name__)


def energy_forces_residual(identifier, natoms, weight, prediction, reference, data):
    r"""A residual function using both energy and forces.

    Parameters
    ----------
    identifier: str
        identifier of the configuration, i.e. path to the file

    natoms: int
        number of atoms in the configuration

    weight: float
        weight for the configuration

    prediction: 1D array
        prediction computed by calculator

    reference: 1D array
        references data for the prediction

    data: dict
        Additional data to be used to calculate the residual. Supported key value pairs
        are:

        - energy_weight: float (default: 1)
        - forces_weight: float (default: 1)
        - normalize_by_atoms: bool (default: True)

    Return
    ------
    residual: 1D array

    Note
    ----
    The length of `prediction` and `reference` (call it `S`) are the same, and it
    depends on `use_energy` and `use_forces` in Calculator. Assume the
    configuration contains of `N` atoms.

    1) If `use_energy == True` and `use_forces == False`, then `S = 1`.
    `prediction[0]` is the potential energy computed by the calculator, and
    `reference[0]` is the reference energy.

    2) If `use_energy == False` and `use_forces == True`, then `S = 3N`.
    `prediction[3*i+0]`, `prediction[3*i+1]`, and `prediction[3*i+2]` are the
    x, y, and z component of the forces on atom i in the configuration, respectively.
    Correspondingly, `reference` is the 3N concatenated reference forces.


    3) If `use_energy == True` and `use_forces == True`, then `S = 3N + 1`.
    `prediction[0]` is the potential energy computed by the calculator, and
    `reference[0]` is the reference energy.
    `prediction[3*i+1]`, `prediction[3*i+2]`, and `prediction[3*i+3]` are the
    x, y, and z component of the forces on atom i in the configuration, respectively.
    Correspondingly, `reference` is the 3N concatenated reference forces.
    """

    # prepare weight based on user provided data
    energy_weight = data['energy_weight']
    forces_weight = data['forces_weight']
    normalize = data['normalize_by_natoms']
    if normalize:
        energy_weight /= natoms
        forces_weight /= natoms

    # obtain residual and properly normalize it
    residual = weight * (prediction - reference)
    residual[0] *= energy_weight
    residual[1:] *= forces_weight

    return residual


def forces_residual(conf_id, natoms, prediction, reference, data):
    data['energy_weight'] = 0
    return energy_forces_residual(conf_id, natoms, prediction, reference, data)


def energy_residual(conf_id, natoms, prediction, reference, data):
    data['forces_weight'] = 0
    return energy_forces_residual(conf_id, natoms, prediction, reference, data)


class Loss(object):
    r"""Objective function class to conduct the optimization."""

    def __new__(
        self, calculator, nprocs=1, residual_fn=energy_forces_residual, residual_data=None
    ):
        """
        Parameters
        ----------

        calculator: Calculator object

        nprocs: int
          Number of processors to parallel to run the predictors.

        residual_fn: function
          function to compute residual, see `energy_forces_residual` for an example

        residual_data: dict
          data passed to residual function

        """
        data = self.check_residual_data(residual_data)
        self.check_computation_flag(calculator, data)

        calc_type = calculator.__class__.__name__

        if 'Torch' in calc_type:
            return LossNeuralNetworkModel(calculator, nprocs, residual_fn, data)
        else:
            return LossPhysicsMotivatedModel(calculator, nprocs, residual_fn, data)

    @staticmethod
    def check_residual_data(data):
        default = {
            'energy_weight': 1.0,
            'forces_weight': 1.0,
            'stress_weight': 1.0,
            'normalize_by_natoms': True,
        }
        if data is not None:
            for key, value in data.items():
                if key not in default:
                    msg = '"{}" not supported by "residual_data".'.format(key)
                    log_entry(logger, msg, level='error')
                    raise InputError(msg)
                else:
                    default[key] = value
        return default

    @staticmethod
    def check_computation_flag(calculator, data):
        ew = data['energy_weight']
        fw = data['forces_weight']
        sw = data['stress_weight']
        msg = (
            '"{0}_weight" set to "{1}". Seems you do not want to use {0} in the fitting. '
            'You can set "use_{0}" of "calculator.create()" to "False" to speed up the '
            'fitting.'
        )

        if calculator.use_energy and ew < 1e-12:
            msg = msg.format('energy', ew)
            log_entry(logger, msg, level='warning')
        if calculator.use_forces and fw < 1e-12:
            msg = msg.format('forces', fw)
            log_entry(logger, msg, level='warning')
        if calculator.use_stress and sw < 1e-12:
            msg = msg.format('stress', sw)
            log_entry(logger, msg, level='warning')


class LossPhysicsMotivatedModel(object):
    r"""Objective function class to conduct the optimization for PM models."""

    scipy_minimize_methods = [
        'Nelder-Mead',
        'Powell',
        'CG',
        'BFGS',
        'Newton-CG',
        'L-BFGS-B',
        'TNC',
        'COBYLA',
        'SLSQP',
        'trust-constr',
        'dogleg',
        'trust-ncg',
        'trust-exact',
        'trust-krylov',
    ]
    scipy_minimize_methods_not_supported_args = ['bounds']
    scipy_least_squares_methods = ['trf', 'dogbox', 'lm', 'geodesiclm']
    scipy_least_squares_methods_not_supported_args = ['bounds']

    def __init__(
        self, calculator, nprocs=1, residual_fn=energy_forces_residual, residual_data=None
    ):

        self.calculator = calculator
        self.nprocs = nprocs

        self.residual_fn = residual_fn
        self.residual_data = residual_data if residual_data is not None else dict()
        self.calculator_type = calculator.__class__.__name__

        if self.calculator_type == 'WrapperCalculator':
            calculators = self.calculator.calculators
        else:
            calculators = [self.calculator]
        for calc in calculators:
            infl_dist = calc.model.get_influence_distance()
            cas = calc.get_compute_arguments()
            for ca in cas:
                ca.refresh(infl_dist)

        logger.info('"{}" instantiated.'.format(self.__class__.__name__))

    def minimize(self, method, **kwargs):
        r"""Minimize the loss.

        Parameters
        ----------
        method: str
            minimization methods as specified at:
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html

        kwargs: dict
            extra keyword arguments that can be used by the scipy optimizer
        """
        kwargs = self.adjust_kwargs(method, **kwargs)

        msg = 'Start minimization using method: {}.'.format(method)
        log_entry(logger, msg, level='info')

        result = self.scipy_optimize(method, **kwargs)

        msg = 'Finish minimization using method: {}.'.format(method)
        log_entry(logger, msg, level='info')

        # update final optimized parameters
        self.calculator.update_opt_params(result.x)

        return result

    def adjust_kwargs(self, method, **kwargs):

        if method in self.scipy_least_squares_methods:

            # check support status
            for i in self.scipy_least_squares_methods_not_supported_args:
                if i in kwargs:
                    msg = (
                        'Argument "{}" should not be set via the "minimize" method. '
                        'It it set internally.'.format(i)
                    )
                    log_entry(logger, msg, level='error')
                    raise LossError(msg)

            # adjust bounds
            if self.calculator.has_opt_params_bounds():
                if method in ['trf', 'dogbox']:
                    bounds = self.calculator.get_opt_params_bounds()
                    lb = [b[0] if b[0] is not None else -np.inf for b in bounds]
                    ub = [b[1] if b[1] is not None else np.inf for b in bounds]
                    bounds = (lb, ub)
                    kwargs['bounds'] = bounds
                else:
                    msg = 'Method "{}" cannot handle bounds.'.format(method)
                    log_entry(logger, msg, level='error')
                    raise LossError(msg)

        elif method in self.scipy_minimize_methods:

            # check support status
            for i in self.scipy_minimize_methods_not_supported_args:
                if i in kwargs:
                    msg = (
                        'Argument "{}" should not be set via the "minimize" method. '
                        'It it set internally.'.format(i)
                    )
                    log_entry(logger, msg, level='error')
                    raise LossError(msg)

            # adjust bounds
            if self.calculator.has_opt_params_bounds():
                if method in ['L-BFGS-B', 'TNC', 'SLSQP']:
                    bounds = self.calculator.get_opt_params_bounds()
                    kwargs['bounds'] = bounds
                else:
                    msg = 'Method "{}" cannot handle bounds.'.format(method)
                    log_entry(logger, msg, level='error')
                    raise LossError(msg)
        else:
            msg = 'minimization method "{}" not supported.'.format(method)
            log_entry(logger, msg, level='error')
            raise LossError(msg)

        return kwargs

    def scipy_optimize(self, method, **kwargs):

        size = parallel.get_MPI_world_size()

        if size > 1:
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()

            msg = 'Running in MPI mode with {} processes.'.format(size)
            log_entry(logger, msg, level='info', print_end='\n\n')

            if self.nprocs > 1:
                msg = (
                    'Argument "nprocs = {}" provided at initialization is ignored. When '
                    'running in MPI mode, the number of processes provided along with '
                    'the "mpiexec" (or "mpirun") command is used.'.format(self.nprocs)
                )
                log_entry(logger, msg, level='warning')

            x = self.calculator.get_opt_params()
            if method in self.scipy_least_squares_methods:
                # geodesic LM
                if method == 'geodesiclm':
                    from geodesicLM import geodesiclm

                    minimize_fn = geodesiclm
                else:
                    minimize_fn = scipy.optimize.least_squares
                func = self.get_residual_MPI
            elif method in self.scipy_minimize_methods:
                minimize_fn = scipy.optimize.minimize
                func = self.get_loss_MPI

            if rank == 0:
                result = minimize_fn(func, x, method=method, **kwargs)
                # notify other process to break func
                break_flag = True
                for i in range(1, size):
                    comm.send(break_flag, dest=i, tag=i)
            else:
                func(x)
                result = None
            result = comm.bcast(result, root=0)
            return result

        else:
            # 1. running MPI with 1 process
            # 2. running without MPI at all
            # both cases are regarded as running without MPI

            if self.nprocs == 1:
                msg = 'Running in serial mode.'
                log_entry(logger, msg, level='info', print_end='\n\n')
            else:
                msg = 'Running in multiprocessing mode with {} processes.'.format(
                    self.nprocs
                )
                log_entry(logger, msg, level='info', print_end='\n\n')

                # Maybe one thinks he is using MPI because nprocs is used
                if mpi4py_available:
                    msg = (
                        '"mpi4y" detected. If you try to run in MPI mode, you should '
                        'execute your code via "mpiexec" (or "mpirun"). If not, ignore '
                        'this message.'
                    )
                    log_entry(logger, msg, level='warning')

            x = self.calculator.get_opt_params()
            if method in self.scipy_least_squares_methods:
                if method == 'geodesiclm':
                    from geodesicLM import geodesiclm

                    minimize_fn = geodesiclm
                else:
                    minimize_fn = scipy.optimize.least_squares
                func = self.get_residual
            elif method in self.scipy_minimize_methods:
                minimize_fn = scipy.optimize.minimize
                func = self.get_loss

            result = minimize_fn(func, x, method=method, **kwargs)
            return result

    def get_residual(self, x):
        r"""Compute the residual.

        This is a callable for optimizing method in scipy.optimize.least_squares,
        which is passed as the first positional argument.

        Parameters
        ----------

        x: 1D array
          optimizing parameter values
        """

        # publish params x to predictor
        self.calculator.update_opt_params(x)

        cas = self.calculator.get_compute_arguments()

        if self.calculator_type == 'WrapperCalculator':
            calc_list = self.calculator.get_calculator_list()
            X = zip(cas, calc_list)
            if self.nprocs > 1:
                residuals = parallel.parmap2(
                    self.get_residual_single_config,
                    X,
                    self.residual_fn,
                    self.residual_data,
                    nprocs=self.nprocs,
                    tuple_X=True,
                )
                residual = np.concatenate(residuals)
            else:
                residual = []
                for ca, calc in X:
                    current_residual = self.get_residual_single_config(
                        ca, calc, self.residual_fn, self.residual_data
                    )
                    residual = np.concatenate((residual, current_residual))

        else:
            if self.nprocs > 1:
                residuals = parallel.parmap2(
                    self.get_residual_single_config,
                    cas,
                    self.calculator,
                    self.residual_fn,
                    self.residual_data,
                    nprocs=self.nprocs,
                    tuple_X=False,
                )
                residual = np.concatenate(residuals)
            else:
                residual = []
                for ca in cas:
                    current_residual = self.get_residual_single_config(
                        ca, self.calculator, self.residual_fn, self.residual_data
                    )
                    residual = np.concatenate((residual, current_residual))

        return residual

    def get_loss(self, x):
        r"""Compute the loss.

        This is a callable for optimizing method in scipy.optimize.minimize,
        which is passed as the first positional argument.

        Parameters
        ----------

        x: 1D array
          optimizing parameter values
        """
        residual = self.get_residual(x)
        loss = 0.5 * np.linalg.norm(residual) ** 2
        return loss

    def get_residual_MPI(self, x):
        def residual_my_chunk(x):
            # broadcast parameters
            x = comm.bcast(x, root=0)
            # publish params x to predictor
            self.calculator.update_opt_params(x)

            residual = []
            for ca in cas:
                current_residual = self.get_residual_single_config(
                    ca, self.calculator, self.residual_fn, self.residual_data
                )
                residual.extend(current_residual)
            return residual

        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        size = comm.Get_size()

        # get my chunk of data
        cas = self.split_data()

        while True:

            if rank == 0:
                break_flag = False
                for i in range(1, size):
                    comm.send(break_flag, dest=i, tag=i)
                residual = residual_my_chunk(x)
                all_residuals = comm.gather(residual, root=0)
                return np.concatenate(all_residuals)
            else:
                break_flag = comm.recv(source=0, tag=rank)
                if break_flag:
                    break
                else:
                    residual = residual_my_chunk(x)
                    all_residuals = comm.gather(residual, root=0)

    def get_loss_MPI(self, x):
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()

        residual = self.get_residual_MPI(x)
        if rank == 0:
            loss = 0.5 * np.linalg.norm(residual) ** 2
        else:
            loss = None

        return loss

    # NOTE this function can be called only once, not need to call each time
    # get_residual_MPI is called
    def split_data(self):
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        size = comm.Get_size()

        # get a portion of data based on rank
        cas = self.calculator.get_compute_arguments()
        # random.shuffle(cas)

        rank_size = len(cas) // size
        # last rank deal with the case where len(cas) cannot evenly divide size
        if rank == size - 1:
            cas = cas[rank_size * rank :]
        else:
            cas = cas[rank_size * rank : rank_size * (rank + 1)]

        return cas

    def get_residual_single_config(self, ca, calculator, residual_fn, residual_data):

        # prediction data
        calculator.compute(ca)
        pred = calculator.get_prediction(ca)

        # reference data
        ref = calculator.get_reference(ca)

        conf = ca.conf
        identifier = conf.get_identifier()
        natoms = conf.get_number_of_atoms()
        weight = conf.get_weight()

        residual = residual_fn(identifier, natoms, weight, pred, ref, residual_data)

        return residual


class LossNeuralNetworkModel(object):
    r"""Objective function class to conduct the optimization for ML models.

    Parameters
    ----------
    calculator: Calculator object

    nprocs: int
      Number of processors to parallel to run the predictors.

    residual_fn: function
      Function to compute residual, see `energy_forces_residual` for an example.

    residual_data: dict
      Data passed to residual function.
    """

    torch_minimize_methods = [
        'Adadelta',
        'Adagrad',
        'Adam',
        'SparseAdam',
        'Adamax',
        'ASGD',
        'LBFGS',
        'RMSprop',
        'Rprop',
        'SGD',
    ]

    def __init__(
        self, calculator, nprocs=1, residual_fn=energy_forces_residual, residual_data=None
    ):

        if not torch_available:
            report_import_error('pytorch')

        self.calculator = calculator
        self.nprocs = nprocs

        self.residual_fn = residual_fn
        self.residual_data = residual_data if residual_data is not None else dict()

        self.calculator_type = calculator.__class__.__name__

        self.optimizer = None
        self.optimizer_stat_path = None

        logger.info('"{}" instantiated.'.format(self.__class__.__name__))

    def minimize(self, method, batch_size=100, num_epochs=1000, start_epoch=0, **kwargs):
        r"""Minimize the loss.

        Parameters
        ----------
        method: str
            PyTorch optimization methods, and available ones are:
            [`Adadelta`, `Adagrad`, `Adam`, `SparseAdam`, `Adamax`, `ASGD`, `LBFGS`,
            `RMSprop`, `Rprop`, `SGD`]
            See also: https://pytorch.org/docs/stable/optim.html

        batch_size: int
            Number of configurations used in in each minimization step.

        num_epochs: int
            Number of epochs to carry out the minimization.

        start_epoch: int
            The starting epoch number. This is typically 0, but if continuing a training,
            it is useful to set this to the last epoch number of the previous training.

        kwargs: dict
            Extra keyword arguments that can be used by the PyTorch optimizer.
        """
        if method not in self.torch_minimize_methods:
            msg = 'Minimization method "{}" not supported.'.format(method)
            log_entry(logger, msg, level='error')
            raise LossError(msg)

        self.method = method
        self.batch_size = batch_size
        self.num_epochs = num_epochs
        self.start_epoch = start_epoch

        # data loader
        loader = self.calculator.get_compute_arguments(batch_size)

        # model save metadata
        save_prefix = self.calculator.model.save_prefix
        save_start = self.calculator.model.save_start
        save_frequency = self.calculator.model.save_frequency
        if save_prefix is None or save_start is None or save_frequency is None:
            logger.info(
                'Model saving meta data not set by user. Now set it to '
                '"prefix=./kliff_saved_model", "start=1", and "frequency=10".'
            )
            save_prefix = os.path.join(os.getcwd(), 'kliff_saved_model')
            save_start = 1
            save_frequency = 10
            self.calculator.model.set_save_metadata(
                save_prefix, save_start, save_frequency
            )

        msg = 'Start minimization using optimization method: {}.'.format(self.method)
        log_entry(logger, msg, level='info')

        # optimizing
        try:
            self.optimizer = getattr(torch.optim, method)(
                self.calculator.model.parameters(), **kwargs
            )
            if self.optimizer_stat_path is not None:
                self._load_optimizer_stat(self.optimizer_stat_path)

        except TypeError as e:
            print(str(e))
            idx = str(e).index("argument '") + 10
            err_arg = str(e)[idx:].strip("'")
            msg = 'Argument "{}" not supported by optimizer "{}".'.format(err_arg, method)
            log_entry(logger, msg, level='error')
            raise InputError(msg)

        for epoch in range(self.start_epoch, self.start_epoch + self.num_epochs):

            # get the loss without any optimization if continue a training
            if self.start_epoch != 0 and epoch == self.start_epoch:
                epoch_loss = self._get_loss_epoch(loader)
                print('Epoch = {:<6d}  loss = {:.10e}'.format(epoch, epoch_loss))

            else:
                epoch_loss = 0
                for ib, batch in enumerate(loader):

                    def closure():
                        self.optimizer.zero_grad()
                        loss = self.get_loss_batch(batch)
                        loss.backward()
                        return loss

                    loss = self.optimizer.step(closure)
                    # float() such that do not accumulate history, more memory friendly
                    epoch_loss += float(loss)

                print('Epoch = {:<6d}  loss = {:.10e}'.format(epoch, epoch_loss))
                if epoch >= save_start and (epoch - save_start) % save_frequency == 0:
                    path = os.path.join(save_prefix, 'model_epoch{}.pkl'.format(epoch))
                    self.calculator.model.save(path)

        # print loss from final parameter and save last epoch
        epoch += 1
        epoch_loss = self._get_loss_epoch(loader)
        print('Epoch = {:<6d}  loss = {:.10e}'.format(epoch, epoch_loss))
        path = os.path.join(save_prefix, 'model_epoch{}.pkl'.format(epoch))
        self.calculator.model.save(path)

        msg = 'Finish minimization using optimization method: {}.'.format(self.method)
        log_entry(logger, msg, level='info')

    def _get_loss_epoch(self, loader):
        epoch_loss = 0
        for ib, batch in enumerate(loader):
            loss = self.get_loss_batch(batch)
            epoch_loss += float(loss)
        return epoch_loss

    def get_loss_batch(self, batch, normalize=True):
        r"""Compute the loss of a batch of samples.

        Parameters
        ----------
        batch: list
            A list of samples.

        normalize: bool
            If `True`, normalize the loss of the batch by the size of the batch.
            Note, how to normalize the loss of a single configuration is determined by the
            `normalize` flag of the `residual_data` argument of :mod:`kliff.Loss`.
        """
        results = self.calculator.compute(batch)
        energy_batch = results['energy']
        forces_batch = results['forces']
        stress_batch = results['stress']

        if forces_batch is None:
            forces_batch = [None] * len(batch)
        if stress_batch is None:
            stress_batch = [None] * len(batch)

        # Instead of loss_batch = 0 and loss_batch += loss in the loop, the below one may
        # be faster, considering chain rule it needs to take derivatives.
        # Anyway, it is minimal. Don't worry about it.
        losses = []
        for sample, energy, forces, stress in zip(
            batch, energy_batch, forces_batch, stress_batch
        ):
            loss = self.get_loss_single_config(sample, energy, forces, stress)
            losses.append(loss)
        loss_batch = torch.stack(losses).sum()
        if normalize:
            loss_batch /= len(batch)

        return loss_batch

    def get_loss_single_config(self, sample, pred_energy, pred_forces, pred_stress):

        if self.calculator.use_energy:
            pred = pred_energy.reshape(-1)  # reshape scalar as 1D tensor
            ref = sample['energy'].reshape(-1)

        if self.calculator.use_forces:
            ref_forces = sample['forces']
            if self.calculator.use_energy:
                pred = torch.cat((pred, pred_forces.reshape(-1)))
                ref = torch.cat((ref, ref_forces.reshape(-1)))
            else:
                pred = pred_forces.reshape(-1)
                ref = ref_forces.reshape(-1)

        if self.calculator.use_stress:
            ref_stress = sample['stress']
            if self.calculator.use_energy or self.calculator.use_stress:
                pred = torch.cat((pred, pred_stress.reshape(-1)))
                ref = torch.cat((ref, ref_stress.reshape(-1)))
            else:
                pred = pred_stress.reshape(-1)
                ref = ref_stress.reshape(-1)

        conf = sample['configuration']
        identifier = conf.get_identifier()
        natoms = conf.get_number_of_atoms()
        weight = conf.get_weight()

        residual = self.residual_fn(
            identifier, natoms, weight, pred, ref, self.residual_data
        )
        loss = torch.sum(torch.pow(residual, 2))

        return loss

    def save_optimizer_stat(self, path='optimizer_stat.pkl'):
        torch.save(self.optimizer.state_dict(), path)

    def load_optimizer_stat(self, path='optimizer_stat.pkl'):
        self.optimizer_stat_path = path

    def _load_optimizer_stat(self, path):
        self.optimizer.load_state_dict(torch.load(path))


class LossError(Exception):
    def __init__(self, msg):
        super(LossError, self).__init__(msg)
        self.msg = msg

    def __expr__(self):
        return self.msg



"""
.. _tut_kim_sw:

Train a Stillinger-Weber potential
==================================

In this tutorial, we train a Stillinger-Weber (SW) potential for silicon that is archived
on OpenKIM_.
"""


##########################################################################################
# Before getting started to train the SW model, let's first install the SW model::
#
#    $ kim-api-collections-management install user SW_StillingerWeber_1985_Si__MO_405512056662_005
#
# .. seealso::
#    This installs the model and its driver into the ``User Collection``. See
#    :ref:`install_model` for more information about installing KIM models.
#
# We are going to create potentials for diamond silicon, and fit the potentials to a
# training set of energies and forces consisting of compressed and stretched diamond
# silicon structures, as well as configurations drawn from molecular dynamics trajectories
# at different temperatures.
# Download the training set :download:`Si_training_set.tar.gz
# <https://raw.githubusercontent.com/mjwen/kliff/master/examples/Si_training_set.tar.gz>`
# and extract the tarball: ``$ tar xzf Si_training_set.tar.gz``. The data is stored in
# **extended xyz** format, and see :ref:`doc.dataset` for more information of this format.
#
# .. warning::
#    The ``Si_training_set`` is just a toy data set for the purpose to demonstrate how to
#    use KLIFF to train potentials. It should not be used to train any potential for real
#    simulations.
#
# Let's first import the modules that will be used in this example.

from kliff.models import KIM
from kliff.loss import Loss
from kliff.calculators import Calculator
from kliff.dataset import Dataset


##########################################################################################
# Model
# -----
#
# We first create a KIM model for the SW potential, and print out all the available
# parameters that can be optimized (we call this ``model parameters``).

model = KIM(model_name='SW_StillingerWeber_1985_Si__MO_405512056662_005')
model.echo_model_params()


##########################################################################################
# The output is generated by the last line, and it tells us the ``name``, ``value``,
# ``size``, ``data type`` and a ``description`` of each parameter.
#
# .. note::
#    You can provide a ``path`` argument to the method ``echo_model_params(path)`` to
#    write the available parameters information to a file indicated by ``path``.
#
# .. note::
#    The available parameters information can also by obtained using the **kliff**
#    :ref:`cmdlntool`:
#    ``$ kliff model --echo-params SW_StillingerWeber_1985_Si__MO_405512056662_005``
#
# Now that we know what parameters are available for fitting, we can optimize all or a
# subset of them to reproduce the training set.

model.set_fitting_params(
    A=[[5.0, 1.0, 20]], B=[['default']], sigma=[[2.0951, 'fix']], gamma=[[1.5]]
)
model.echo_fitting_params()


##########################################################################################
# Here, we tell KLIFF to fit four parameters ``B``, ``gamma``, ``sigma``, and ``A`` of the
# SW model. The information for each fitting parameter should be provided as a list of
# list, where the size of the outer list should be equal to the ``size`` of the parameter
# given by ``model.echo_model_params()``. For each inner list, you can provide either one,
# two, or three items.
#
# - One item. You can use a numerical value (e.g. ``gamma``) to provide an initial guess
#   of the parameter. Alternatively, the string ``'default'`` can be provided to use the
#   default value in the model (e.g. ``B``).
#
# - Two items. The first item should be a numerical value and the second item should be
#   the string ``'fix'`` (e.g. ``sigma``), which tells KLIFF to use the value for the
#   parameter, but do not optimize it.
#
# - Three items. The first item can be a numerical value or the string ``'default'``,
#   having the same meanings as the one item case. In the second and third items, you can
#   list the lower and upper bounds for the parameters, respectively. A bound could be
#   provided as a numerical values or ``None``. The latter indicates no bound is applied.
#
# The call of ``model.echo_fitting_params()`` prints out the fitting parameters that we
# require KLIFF to optimize. The number ``1`` after the name of each parameter indicates
# the size of the parameter.
#
# .. note::
#    The parameters that are not included as a fitting parameter are fixed to the default
#    values in the model during the optimization.
#
#
# Training set
# ------------
#
# KLIFF has a :class:`~kliff.dataset.Dataset` to deal with the training data (and possibly
# test data). For the silicon training set, we can read and process the files by:

dataset_name = 'Si_training_set'
tset = Dataset()
tset.read(dataset_name)
configs = tset.get_configs()


##########################################################################################
# The ``configs`` in the last line is a list of :class:`~kliff.dataset.Configuration`.
# Each configuration is an internal representation of a processed **extended xyz** file,
# hosting the species, coordinates, energy, forces, and other related information of a
# system of atoms.
#
#
# Calculator
# ----------
#
# :class:`~kliff.calculator.Calculator` is the central agent that exchanges information
# and orchestrate the operation of the fitting process. It calls the model to compute the
# energy and forces and provide this information to the `Loss function`_ (discussed below)
# to compute the loss. It also grabs the parameters from the optimizer and update the
# parameters stored in the model so that the up-to-date parameters are used the next time
# the model is evaluated to compute the energy and forces. The calculator can be created
# by:

calc = Calculator(model)
calc.create(configs)


##########################################################################################
# where ``calc.create(configs)`` does some initializations for each
# configuration in the training set, such as creating the neighbor list.
#
#
# Loss function
# -------------
#
# KLIFF uses a loss function to quantify the difference between the training set data and
# potential predictions and uses minimization algorithms to reduce the loss as much as
# possible. KLIFF provides a large number of minimization algorithms by interacting with
# SciPy_. For physics-motivated potentials, any algorithm listed on
# `scipy.optimize.minimize`_ and `scipy.optimize.least_squares`_ can be used. In the
# following code snippet, we create a loss of energy and forces, where the residual
# function uses an ``energy_weight`` of ``1.0`` and a ``forces_weight`` of ``0.1``, and
# ``2`` processors will be used to calculate the loss. The ``L-BFGS-B`` minimization
# algorithm is applied to minimize the loss, and the minimization is allowed to run for
# a max number of 100 iterations.

steps = 100
residual_data = {'energy_weight': 1.0, 'forces_weight': 0.1}
loss = Loss(calc, residual_data=residual_data, nprocs=2)
loss.minimize(method='L-BFGS-B', options={'disp': True, 'maxiter': steps})


##########################################################################################
# The minimization stops after running for 27 steps.  After the minimization, we'd better
# save the model, which can be loaded later for the purpose to do a retraining or
# evaluations. If satisfied with the fitted model, you can also write it as a KIM model
# that can be used with LAMMPS_, GULP_, ASE_, etc. via the kim-api_.

model.echo_fitting_params()
model.save('kliff_model.pkl')
model.write_kim_model()
model.load('kliff_model.pkl')


##########################################################################################
# The first line of the above code generates the output.  A comparison with the original
# parameters before carrying out the minimization shows that we recover the original
# parameters quite reasonably. The second line saves the fitted model to a file named
# ``kliff_model.pkl`` on the disk, and the third line writes out a KIM potential named
# ``SW_StillingerWeber_1985_Si__MO_405512056662_005_kliff_trained``.
#
# .. seealso::
#    For information about how to load a saved model, see :ref:`doc.modules`.
#
#
# .. _OpenKIM: https://openkim.org
# .. _SciPy: https://scipy.org
# .. _scipy.optimize.minimize: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
# .. _scipy.optimize.least_squares: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html
# .. _kim-api: https://openkim.org/kim-api/
# .. _LAMMPS: https://lammps.sandia.gov
# .. _GULP: http://gulp.curtin.edu.au/gulp/
# .. _ASE: https://wiki.fysik.dtu.dk/ase/



import numpy as np
import collections
import os


class WrapCalculator:
    """ Wrapper over user-defined predictor function.

    Parameters
    ----------

    runner, function
      User defined function to generate prediction for a material property.
      For example, using LAMMPS to compute the elastic constant.

    outname, str
      Name of file that stores the results generated by 'runner'. The file should
      be in EDN format.

    keys: list of str
      Keywords in the 'outname' EDN file, whose value will be parsed as the prediction.
    """

    def __init__(self, params, outname, keys, runner, *args):
        self.params = params
        self.outname = outname
        self.keys = keys
        self.runner = runner
        self.args = args

    def get_prediction(self):
        """
        Return 1D array of floats.
        """
        self._update_params()
        self.runner(*self.args)
        return self._parse_edn(self.outname, self.keys)

    def _update_params(self):
        """Write parameters to file KIM_MODEL_PARAMS + modelname, and also give its path to the
        enviroment variable that has the same name: KIM_MODEL_PARAMS + modelname.
        """
        name = 'KIM_MODEL_PARAMS_' + self.params._modelname
        self.params.echo_params(name, print_size=True)
        path = os.getcwd() + os.path.sep + name
        os.environ[name] = path

    def _parse_edn(self, fname, keys):
        """ Wrapper to use end_format to parse output file of 'runner' in edn format.

        Parameters
        ----------

        fname: str
          Name of the output file of OpenKIM test.

        keys: list of str
          Keyword in the edn format(think it as a dictionary), whose value will be returned.
        """

        with open(fname, 'r') as fin:
            lines = fin.read()
        parsed = edn_format.loads(lines)
        values = []
        for k in keys:
            try:
                v = parsed[k]['source-value']
            except KeyError:
                raise KeyError('Keyword "{}" not found in {}.'.format(k, self.outname))
            # make it a 1D array
            # we expect v as a python built-in object (if numpy object, this will fail)
            if isinstance(v, collections.Sequence):
                v = np.array(v).flatten()
            else:
                v = [v]
            values.append(v)
        return np.concatenate(values)
